from unittest.mock import patch
from urllib.parse import urlparse

from murfey.client.context import ensure_dcg_exists
from murfey.client.instance_environment import MurfeyInstanceEnvironment


@patch("murfey.client.context.capture_post")
def test_ensure_dcg_exists_tomo(mock_capture_post, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path / "cm12345-6/metadata_folder"],
        default_destinations={
            tmp_path
            / "cm12345-6/metadata_folder": f"{tmp_path}/destination/cm12345-6/raw"
        },
        instrument_name="",
        visit="cm12345-6",
        murfey_session=1,
    )

    metadata_source = tmp_path / "cm12345-6/metadata_folder"
    metadata_source.mkdir(parents=True)
    with open(metadata_source / "Session.dm", "w") as dm_file:
        dm_file.write(
            "<TomographySession><AtlasId>"
            r"X:\cm12345-6\atlas\atlas_metadata\Sample6\Atlas\Atlas.dm"
            "</AtlasId></TomographySession>"
        )

    atlas_xml = tmp_path / "cm12345-6/atlas/atlas_metadata/Sample6/Atlas/Atlas_4.xml"
    atlas_xml.parent.mkdir(parents=True)
    with open(atlas_xml, "w") as xml_file:
        xml_file.write(
            "<MicroscopeImage><SpatialScale><pixelSize><x><numericValue>4.7"
            "</numericValue></x></pixelSize></SpatialScale></MicroscopeImage>"
        )

    ensure_dcg_exists(
        collection_type="tomo",
        metadata_source=metadata_source,
        environment=env,
        token="token",
    )

    dcg_data = {
        "experiment_type_id": 36,
        "tag": f"{tmp_path}/metadata_folder",
        "atlas": f"{tmp_path}/destination/{atlas_xml.relative_to(tmp_path).with_suffix('.jpg')}",
        "sample": 6,
        "atlas_pixel_size": 4.7 * 7.8,
    }
    mock_capture_post.assert_called_once_with(
        base_url="http://localhost:8000",
        router_name="workflow.router",
        function_name="register_dc_group",
        token="token",
        visit_name="cm12345-6",
        session_id=1,
        data=dcg_data,
    )


@patch("murfey.client.context.capture_post")
def test_ensure_dcg_exists_spa(mock_capture_post, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path / "cm12345-6/metadata_folder"],
        default_destinations={
            tmp_path
            / "cm12345-6/metadata_folder": f"{tmp_path}/destination/cm12345-6/raw",
        },
        instrument_name="",
        visit="cm12345-6",
        murfey_session=1,
    )

    metadata_source = tmp_path / "cm12345-6/metadata_folder"
    metadata_source.mkdir(parents=True)
    with open(metadata_source / "EpuSession.dm", "w") as dm_file:
        dm_file.write(
            "<EpuSessionXml><Samples><_items><SampleXml><AtlasId z:Id='10'>"
            r"X:\cm12345-6\atlas\atlas_metadata\Sample6\Atlas\Atlas.dm"
            "</AtlasId></SampleXml><SampleXml></SampleXml></_items></Samples></EpuSessionXml>"
        )

    # Make data location
    (tmp_path / "metadata_folder/Images-Disc1").mkdir(parents=True)

    atlas_xml = tmp_path / "cm12345-6/atlas/atlas_metadata/Sample6/Atlas/Atlas_4.xml"
    atlas_xml.parent.mkdir(parents=True)
    with open(atlas_xml, "w") as xml_file:
        xml_file.write(
            "<MicroscopeImage><SpatialScale><pixelSize><x><numericValue>4.7"
            "</numericValue></x></pixelSize></SpatialScale></MicroscopeImage>"
        )

    ensure_dcg_exists(
        collection_type="spa",
        metadata_source=metadata_source / "Images-Disc1",
        environment=env,
        token="token",
    )

    dcg_data = {
        "experiment_type_id": 37,
        "tag": f"{tmp_path}/metadata_folder/Images-Disc1",
        "atlas": f"{tmp_path}/destination/{atlas_xml.relative_to(tmp_path).with_suffix('.jpg')}",
        "sample": 6,
        "atlas_pixel_size": 4.7 * 7.8,
    }
    mock_capture_post.assert_called_once_with(
        base_url="http://localhost:8000",
        router_name="workflow.router",
        function_name="register_dc_group",
        token="token",
        visit_name="cm12345-6",
        session_id=1,
        data=dcg_data,
    )


@patch("murfey.client.context.capture_post")
def test_ensure_dcg_exists_spa_missing_xml(mock_capture_post, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path],
        default_destinations={tmp_path: str(tmp_path)},
        instrument_name="",
        visit="cm12345-6",
        murfey_session=1,
    )

    metadata_source = tmp_path / "cm12345-6/metadata_folder"
    ensure_dcg_exists(
        collection_type="spa",
        metadata_source=metadata_source,
        environment=env,
        token="token",
    )

    dcg_data = {
        "experiment_type_id": 37,
        "tag": f"{tmp_path}/metadata_folder",
    }
    mock_capture_post.assert_called_once_with(
        base_url="http://localhost:8000",
        router_name="workflow.router",
        function_name="register_dc_group",
        token="token",
        visit_name="cm12345-6",
        session_id=1,
        data=dcg_data,
    )



import sys

from pytest_mock import MockerFixture

from murfey.cli.generate_route_manifest import run


def test_run(
    mocker: MockerFixture,
):
    # Mock out print() and exit()
    mock_print = mocker.patch("builtins.print")
    mock_exit = mocker.patch("builtins.exit")

    # Run the function with its args
    sys.argv = ["", "--debug"]
    run()

    # Check that the final print message and exit() are called
    print_calls = mock_print.call_args_list
    last_print_call = print_calls[-1]
    last_printed = last_print_call.args[0]
    assert last_printed.startswith(
        "Route manifest for instrument and backend servers saved to"
    )
    mock_exit.assert_called_once()



import logging
import os
from pathlib import Path
from typing import Dict, List

from sqlmodel import select
from sqlmodel.orm.session import Session as SQLModelSession
from werkzeug.utils import secure_filename

import murfey.server.prometheus as prom
from murfey.util import safe_run, sanitise, secure_path
from murfey.util.config import get_machine_config
from murfey.util.db import (
    DataCollection,
    DataCollectionGroup,
    FoilHole,
    GridSquare,
    ProcessingJob,
    RsyncInstance,
    Session as MurfeySession,
)

logger = logging.getLogger("murfey.server.api.shared")


def remove_session_by_id(session_id: int, db):
    session = db.exec(select(MurfeySession).where(MurfeySession.id == session_id)).one()
    sessions_for_visit = db.exec(
        select(MurfeySession).where(MurfeySession.visit == session.visit)
    ).all()
    # Don't remove prometheus metrics if there are other sessions using them
    if len(sessions_for_visit) == 1:
        safe_run(
            prom.monitoring_switch.remove,
            args=(session.visit,),
            label="monitoring_switch",
        )
        rsync_instances = db.exec(
            select(RsyncInstance).where(RsyncInstance.session_id == session_id)
        ).all()
        for ri in rsync_instances:
            safe_run(
                prom.seen_files.remove,
                args=(ri.source, session.visit),
                label="seen_files",
            )
            safe_run(
                prom.transferred_files.remove,
                args=(ri.source, session.visit),
                label="transferred_files",
            )
            safe_run(
                prom.transferred_files_bytes.remove,
                args=(ri.source, session.visit),
                label="transferred_files_bytes",
            )
            safe_run(
                prom.seen_data_files.remove,
                args=(ri.source, session.visit),
                label="seen_data_files",
            )
            safe_run(
                prom.transferred_data_files.remove,
                args=(ri.source, session.visit),
                label="transferred_data_files",
            )
            safe_run(
                prom.transferred_data_files_bytes.remove,
                args=(ri.source, session.visit),
                label="transferred_data_file_bytes",
            )
            safe_run(
                prom.skipped_files.remove,
                args=(ri.source, session.visit),
                label="skipped_files",
            )
    collected_ids = db.exec(
        select(DataCollectionGroup, DataCollection, ProcessingJob)
        .where(DataCollectionGroup.session_id == session_id)
        .where(DataCollection.dcg_id == DataCollectionGroup.id)
        .where(ProcessingJob.dc_id == DataCollection.id)
    ).all()
    for c in collected_ids:
        safe_run(
            prom.preprocessed_movies.remove,
            args=(c[2].id,),
            label="preprocessed_movies",
        )
    db.delete(session)
    db.commit()
    logger.debug(f"Successfully removed session {session_id} from database")
    return


def get_grid_squares(session_id: int, db):
    grid_squares = db.exec(
        select(GridSquare).where(GridSquare.session_id == session_id)
    ).all()
    tags = {gs.tag for gs in grid_squares}
    res = {}
    for t in tags:
        res[t] = [gs for gs in grid_squares if gs.tag == t]
    return res


def get_grid_squares_from_dcg(session_id: int, dcgid: int, db) -> List[GridSquare]:
    grid_squares = db.exec(
        select(GridSquare, DataCollectionGroup)
        .where(GridSquare.session_id == session_id)
        .where(GridSquare.tag == DataCollectionGroup.tag)
        .where(DataCollectionGroup.id == dcgid)
    ).all()
    return [gs[0] for gs in grid_squares]


def get_foil_holes_from_grid_square(
    session_id: int, dcgid: int, gsid: int, db
) -> List[FoilHole]:
    foil_holes = db.exec(
        select(FoilHole, GridSquare, DataCollectionGroup)
        .where(FoilHole.grid_square_id == GridSquare.id)
        .where(GridSquare.name == gsid)
        .where(GridSquare.session_id == session_id)
        .where(GridSquare.tag == DataCollectionGroup.tag)
        .where(DataCollectionGroup.id == dcgid)
    ).all()
    return [fh[0] for fh in foil_holes]


def get_foil_hole(session_id: int, fh_name: int, db) -> Dict[str, int]:
    foil_holes = db.exec(
        select(FoilHole, GridSquare)
        .where(FoilHole.name == fh_name)
        .where(FoilHole.session_id == session_id)
        .where(GridSquare.id == FoilHole.grid_square_id)
    ).all()
    return {f[1].tag: f[0].id for f in foil_holes}


def find_upstream_visits(session_id: int, db: SQLModelSession, max_depth: int = 2):
    """
    Returns a nested dictionary, in which visits and the full paths to their directories
    are further grouped by instrument name.
    """

    def _recursive_search(
        dirpath: str | Path,
        search_string: str,
        partial_match: bool = True,
        max_depth: int = 1,
        result: dict[str, Path] | None = None,
    ):
        # If no dictionary was passed in, create a new dictionary
        if result is None:
            result = {}
        # Stop recursing for this route once max depth hits 0
        if max_depth == 0:
            return result

        # Walk through the directories
        try:
            for entry in os.scandir(dirpath):
                if entry.is_dir():
                    # Update dictionary with match and stop recursing for this route
                    if (
                        search_string in entry.name
                        if partial_match
                        else search_string == entry.name
                    ):
                        if result is not None:  # MyPy needs this 'is not None' check
                            result[entry.name] = Path(entry.path)
                    else:
                        # Continue searching down this route until max depth is reached
                        result = _recursive_search(
                            dirpath=entry.path,
                            search_string=search_string,
                            partial_match=partial_match,
                            max_depth=max_depth - 1,
                            result=result,
                        )
        # Safely skip the directory if it can't access it
        except PermissionError:
            logger.warning(f"Unable to access {dirpath}; skipping", exc_info=True)
        return result

    murfey_session = db.exec(
        select(MurfeySession).where(MurfeySession.id == session_id)
    ).one()
    visit_name = murfey_session.visit
    instrument_name = murfey_session.instrument_name
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    upstream_visits: dict[str, dict[str, Path]] = {}
    # Iterates through provided upstream directories
    for (
        upstream_instrument,
        upstream_data_dir,
    ) in machine_config.upstream_data_directories.items():
        # Recursively look for matching visit names under current directory
        upstream_visits[upstream_instrument] = _recursive_search(
            dirpath=upstream_data_dir,
            search_string=f"{visit_name.split('-')[0]}-",
            partial_match=True,
            max_depth=max_depth,
        )
    return upstream_visits


def gather_upstream_files(
    session_id: int,
    upstream_instrument: str,
    upstream_visit_path: Path,
    db: SQLModelSession,
):
    """
    Searches the specified upstream instrument for files based on the search strings
    set in the MachineConfig and returns them as a list of file paths.
    """
    # Load the current instrument's machine config
    murfey_session = db.exec(
        select(MurfeySession).where(MurfeySession.id == session_id)
    ).one()
    instrument_name = murfey_session.instrument_name
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]

    # Search for files using the configured strings for that upstream instrument
    file_list: list[Path] = []
    logger.info(f"Searching for files in {sanitise(str(upstream_visit_path))!r}")
    if (
        machine_config.upstream_data_search_strings.get(upstream_instrument, None)
        is not None
    ):
        for search_string in machine_config.upstream_data_search_strings[
            upstream_instrument
        ]:
            logger.info(f"Using search string {search_string}")
            for file in upstream_visit_path.glob(search_string):
                if file.is_file():
                    file_list.append(file)
        logger.info(
            f"Found {len(file_list)} files for download "
            f"from {sanitise(upstream_instrument)}"
        )
    else:
        logger.warning(
            "Upstream file searching has not been configured for "
            f"{sanitise(upstream_instrument)} on {sanitise(instrument_name)}"
        )
    return file_list


def get_upstream_file(file_path: str | Path):
    file_path = Path(file_path) if isinstance(file_path, str) else file_path
    file_path = secure_path(file_path)
    if file_path.exists() and file_path.is_file():
        return file_path
    logger.warning(f"Requested file {sanitise(str(file_path))!r} was not found")
    return None


def get_upstream_tiff_dirs(visit_name: str, instrument_name: str) -> List[Path]:
    tiff_dirs = []
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    for directory_name in machine_config.upstream_data_tiff_locations:
        for _, p in machine_config.upstream_data_directories.items():
            if (Path(p) / secure_filename(visit_name)).is_dir():
                processed_dir = Path(p) / secure_filename(visit_name) / directory_name
                tiff_dirs.append(processed_dir)
                break
    if not tiff_dirs:
        logger.warning(
            f"No candidate directory found for upstream download from visit {sanitise(visit_name)}"
        )
    return tiff_dirs


def gather_upstream_tiffs(visit_name: str, session_id: int, db: SQLModelSession):
    """
    Looks for TIFF files associated with the current session in the permitted storage
    servers, and returns their relative file paths as a list.
    """
    instrument_name = (
        db.exec(select(MurfeySession).where(MurfeySession.id == session_id))
        .one()
        .instrument_name
    )
    upstream_tiff_paths = []
    tiff_dirs = get_upstream_tiff_dirs(visit_name, instrument_name)
    if not tiff_dirs:
        return None
    for tiff_dir in tiff_dirs:
        for f in tiff_dir.glob("**/*.tiff"):
            upstream_tiff_paths.append(str(f.relative_to(tiff_dir)))
        for f in tiff_dir.glob("**/*.tif"):
            upstream_tiff_paths.append(str(f.relative_to(tiff_dir)))
    return upstream_tiff_paths


def get_tiff_file(
    visit_name: str, session_id: int, tiff_path: str, db: SQLModelSession
):
    instrument_name = (
        db.exec(select(MurfeySession).where(MurfeySession.id == session_id))
        .one()
        .instrument_name
    )
    tiff_dirs = get_upstream_tiff_dirs(visit_name, instrument_name)
    if not tiff_dirs:
        return None

    tiff_path = "/".join(secure_filename(p) for p in tiff_path.split("/"))
    for tiff_dir in tiff_dirs:
        tiff_file = tiff_dir / tiff_path
        if tiff_file.is_file():
            break
    else:
        logger.warning(f"TIFF {tiff_path} not found")
        return None

    return tiff_file



import pytest

from murfey.util.api import url_path_for

url_path_test_matrix: tuple[tuple[str, str, dict[str, str | int], str], ...] = (
    # Router name | Function name | kwargs | Expected URL
    (
        "instrument_server.api.router",
        "health",
        {},
        "/health",
    ),
    (
        "instrument_server.api.router",
        "stop_multigrid_watcher",
        {"session_id": 0, "label": "some_label"},
        "/sessions/0/multigrid_watcher/some_label",
    ),
    (
        "api.hub.router",
        "get_instrument_image",
        {"instrument_name": "test"},
        "/instrument/test/image",
    ),
    (
        "api.instrument.router",
        "check_if_session_is_active",
        {
            "instrument_name": "test",
            "session_id": 0,
        },
        "/instrument_server/instruments/test/sessions/0/active",
    ),
)


@pytest.mark.parametrize("test_params", url_path_test_matrix)
def test_url_path_for(test_params: tuple[str, str, dict[str, str | int], str]):
    # Unpack test params
    router_name, function_name, kwargs, expected_url_path = test_params
    assert (
        url_path_for(router_name=router_name, function_name=function_name, **kwargs)
        == expected_url_path
    )



"""
Functions to process the requests received by Murfey related to the CLEM workflow.

The CLEM-related file registration API endpoints can eventually be moved here, since
the file registration processes all take place on the server side only.
"""

from __future__ import annotations

import json
import logging
import re
import traceback
from collections.abc import Collection
from importlib.metadata import entry_points
from pathlib import Path
from typing import Literal, Optional

from pydantic import BaseModel
from sqlmodel import Session, select

import murfey.util.db as MurfeyDB
from murfey.server import _transport_object
from murfey.util.models import GridSquareParameters
from murfey.util.processing_params import (
    default_clem_processing_parameters as processing_params,
)
from murfey.workflows.clem import get_db_entry
from murfey.workflows.clem.align_and_merge import submit_cluster_request

logger = logging.getLogger("murfey.workflows.clem.register_preprocessing_results")


class CLEMPreprocessingResult(BaseModel):
    series_name: str
    number_of_members: int
    is_stack: bool
    is_montage: bool
    output_files: dict[
        Literal["gray", "red", "green", "blue", "cyan", "magenta", "yellow"], Path
    ]
    thumbnails: dict[
        Literal["gray", "red", "green", "blue", "cyan", "magenta", "yellow"], Path
    ] = {}
    thumbnail_size: Optional[tuple[int, int]] = None  # height, width
    metadata: Path
    parent_lif: Optional[Path] = None
    parent_tiffs: dict[
        Literal["gray", "red", "green", "blue", "cyan", "magenta", "yellow"], list[Path]
    ] = {}
    pixels_x: int
    pixels_y: int
    units: str
    pixel_size: float
    resolution: float
    extent: list[float]  # [x0, x1, y0, y1]


def _is_clem_atlas(result: CLEMPreprocessingResult):
    # If an image has a width/height of at least 1.5 mm, it should qualify as an atlas
    return (
        max(
            result.pixels_x * result.pixel_size,
            result.pixels_y * result.pixel_size,
        )
        >= processing_params.atlas_threshold
    )


COLOR_FLAGS_MURFEY = {
    "gray": "has_grey",
    "red": "has_red",
    "green": "has_green",
    "blue": "has_blue",
    "cyan": "has_cyan",
    "magenta": "has_magenta",
    "yellow": "has_yellow",
}


def _get_color_flags(
    colors: Collection[str] | None = None,
):
    colors = colors or []
    color_flags = dict.fromkeys(COLOR_FLAGS_MURFEY.values(), False)
    for color in colors:
        color_flags[COLOR_FLAGS_MURFEY[color]] = True
    return color_flags


def _register_clem_image_series(
    session_id: int,
    result: CLEMPreprocessingResult,
    murfey_db: Session,
):
    clem_img_series: MurfeyDB.CLEMImageSeries = get_db_entry(
        db=murfey_db,
        table=MurfeyDB.CLEMImageSeries,
        session_id=session_id,
        series_name=result.series_name,
    )
    clem_metadata: MurfeyDB.CLEMImageMetadata = get_db_entry(
        db=murfey_db,
        table=MurfeyDB.CLEMImageMetadata,
        session_id=session_id,
        file_path=result.metadata,
    )
    # Register and link parent LIF file if present
    if result.parent_lif is not None:
        clem_lif_file: MurfeyDB.CLEMLIFFile = get_db_entry(
            db=murfey_db,
            table=MurfeyDB.CLEMLIFFile,
            session_id=session_id,
            file_path=result.parent_lif,
        )
        clem_img_series.parent_lif = clem_lif_file
        clem_metadata.parent_lif = clem_lif_file

    # Link and commit series and metadata tables
    clem_img_series.associated_metadata = clem_metadata
    murfey_db.add_all([clem_img_series, clem_metadata])
    murfey_db.commit()

    # Iteratively register the output image stacks
    for c, (channel, output_file) in enumerate(result.output_files.items()):
        clem_img_stk: MurfeyDB.CLEMImageStack = get_db_entry(
            db=murfey_db,
            table=MurfeyDB.CLEMImageStack,
            session_id=session_id,
            file_path=output_file,
        )

        # Link associated metadata
        clem_img_stk.associated_metadata = clem_metadata
        clem_img_stk.parent_series = clem_img_series
        clem_img_stk.channel_name = channel
        if result.parent_lif is not None:
            clem_img_stk.parent_lif = clem_lif_file
        murfey_db.add(clem_img_stk)
        murfey_db.commit()

        # Register and link parent TIFF files if present
        if result.parent_tiffs:
            seed_file = result.parent_tiffs[channel][0]
            if c == 0:
                # Load list of files to register from seed file
                series_identifier = seed_file.stem.split("--")[0] + "--"
                tiff_list = list(seed_file.parent.glob(f"{series_identifier}--"))

            # Load TIFF files by colour channel if "--C" in file stem
            match = re.search(r"--C[\d]{2,3}", seed_file.stem)
            tiff_file_subset = [
                file
                for file in tiff_list
                if file.stem.startswith(series_identifier)
                and (match.group(0) in file.stem if match else True)
            ]
            tiff_file_subset.sort()

            # Register TIFF file subset
            clem_tiff_files = []
            for file in tiff_file_subset:
                clem_tiff_file: MurfeyDB.CLEMTIFFFile = get_db_entry(
                    db=murfey_db,
                    table=MurfeyDB.CLEMTIFFFile,
                    session_id=session_id,
                    file_path=file,
                )

                # Link associated metadata
                clem_tiff_file.associated_metadata = clem_metadata
                clem_tiff_file.child_series = clem_img_series
                clem_tiff_file.child_stack = clem_img_stk

                clem_tiff_files.append(clem_tiff_file)

            murfey_db.add_all(clem_tiff_files)
            murfey_db.commit()

    # Add metadata for this series
    clem_img_series.image_search_string = str(output_file.parent / "*tiff")
    clem_img_series.data_type = "atlas" if _is_clem_atlas(result) else "grid_square"
    clem_img_series.number_of_members = result.number_of_members
    for col_name, value in _get_color_flags(result.output_files.keys()).items():
        setattr(clem_img_series, col_name, value)
    clem_img_series.collection_mode = _determine_collection_mode(
        result.output_files.keys()
    )
    clem_img_series.image_pixels_x = result.pixels_x
    clem_img_series.image_pixels_y = result.pixels_y
    clem_img_series.image_pixel_size = result.pixel_size
    clem_img_series.units = result.units
    clem_img_series.x0 = result.extent[0]
    clem_img_series.x1 = result.extent[1]
    clem_img_series.y0 = result.extent[2]
    clem_img_series.y1 = result.extent[3]
    # Register thumbnails if they are present
    if result.thumbnails and result.thumbnail_size:
        thumbnail = list(result.thumbnails.values())[0]
        clem_img_series.thumbnail_search_string = str(thumbnail.parent / "*.png")

        thumbnail_height, thumbnail_width = result.thumbnail_size
        scaling_factor = min(
            thumbnail_height / result.pixels_y, thumbnail_width / result.pixels_x
        )
        clem_img_series.thumbnail_pixel_size = result.pixel_size / scaling_factor
        clem_img_series.thumbnail_pixels_x = int(result.pixels_x * scaling_factor)
        clem_img_series.thumbnail_pixels_y = int(result.pixels_y * scaling_factor)
    murfey_db.add(clem_img_series)
    murfey_db.commit()
    murfey_db.close()

    logger.info(f"CLEM preprocessing results registered for {result.series_name!r} ")


def _determine_collection_mode(
    colors: Collection[str] | None = None,
):
    if not colors:
        logger.warning("No colours were present in returned result")
        return None
    if "gray" in colors:
        if len(colors) == 1:
            return "Bright Field"
        else:
            return "Bright Field and Fluorescent"
    else:
        return "Fluorescent"


def _snake_to_camel_case(string: str):
    parts = string.split("_")
    return parts[0] + "".join(part.capitalize() for part in parts[1:])


COLOR_FLAGS_MURFEY_TO_ISPYB = {
    value: _snake_to_camel_case(value) for value in COLOR_FLAGS_MURFEY.values()
}


def _register_dcg_and_atlas(
    session_id: int,
    instrument_name: str,
    visit_name: str,
    result: CLEMPreprocessingResult,
    murfey_db: Session,
):
    # Determine variables to register data collection group and atlas with
    proposal_code = "".join(char for char in visit_name.split("-")[0] if char.isalpha())
    proposal_number = "".join(
        char for char in visit_name.split("-")[0] if char.isdigit()
    )
    visit_number = visit_name.split("-")[-1]

    # Generate name/tag for data colleciton group based on series name
    dcg_name = result.series_name.split("--")[0]
    if result.series_name.split("--")[1].isdigit():
        dcg_name += f"--{result.series_name.split('--')[1]}"

    # Determine values for atlas
    if _is_clem_atlas(result):
        output_file = list(result.output_files.values())[0]
        # Register the thumbnail entries if they are provided
        if result.thumbnails and result.thumbnail_size is not None:
            # Glob path to the thumbnail files
            thumbnail = list(result.thumbnails.values())[0]
            atlas_name = str(thumbnail.parent / "*.png")

            # Work out the scaling factor used
            thumbnail_height, thumbnail_width = result.thumbnail_size
            scaling_factor = min(
                thumbnail_width / result.pixels_x,
                thumbnail_height / result.pixels_y,
            )
            atlas_pixel_size = result.pixel_size / scaling_factor
        # Otherwise, register the TIFF files themselves
        else:
            atlas_name = str(output_file.parent / "*.tiff")
            atlas_pixel_size = result.pixel_size
        # Translate colour flags into ISPyB convention
        color_flags = {
            COLOR_FLAGS_MURFEY_TO_ISPYB[key]: int(value)
            for key, value in _get_color_flags(result.output_files.keys()).items()
        }
        collection_mode = _determine_collection_mode(result.output_files.keys())
    else:
        atlas_name = ""
        atlas_pixel_size = 0.0
        color_flags = None
        collection_mode = None

    if dcg_search := murfey_db.exec(
        select(MurfeyDB.DataCollectionGroup)
        .where(MurfeyDB.DataCollectionGroup.session_id == session_id)
        .where(MurfeyDB.DataCollectionGroup.tag == dcg_name)
    ).all():
        dcg_entry = dcg_search[0]
        # Update atlas if registering atlas dataset
        # and data collection group already exists
        if _is_clem_atlas(result):
            atlas_message = {
                "session_id": session_id,
                "dcgid": dcg_entry.id,
                "atlas_id": dcg_entry.atlas_id,
                "atlas": atlas_name,
                "atlas_pixel_size": atlas_pixel_size,
                "sample": dcg_entry.sample,
                "color_flags": color_flags,
                "collection_mode": collection_mode,
            }
            if entry_point_result := entry_points(
                group="murfey.workflows", name="atlas_update"
            ):
                (workflow,) = entry_point_result
                _ = workflow.load()(
                    message=atlas_message,
                    murfey_db=murfey_db,
                )
            else:
                logger.warning("No workflow found for 'atlas_update'")
    else:
        # Register data collection group and placeholder for the atlas
        dcg_message = {
            "microscope": instrument_name,
            "proposal_code": proposal_code,
            "proposal_number": proposal_number,
            "visit_number": visit_number,
            "session_id": session_id,
            "tag": dcg_name,
            "experiment_type_id": 45,
            "atlas": atlas_name,
            "atlas_pixel_size": atlas_pixel_size,
            "sample": None,
            "color_flags": color_flags,
            "collection_mode": collection_mode,
        }
        if entry_point_result := entry_points(
            group="murfey.workflows", name="data_collection_group"
        ):
            (workflow,) = entry_point_result
            # Register grid square
            _ = workflow.load()(
                message=dcg_message,
                murfey_db=murfey_db,
            )
        else:
            logger.warning("No workflow found for 'data_collection_group'")

    # Store data collection group id in CLEM image series table
    dcg_entry = murfey_db.exec(
        select(MurfeyDB.DataCollectionGroup)
        .where(MurfeyDB.DataCollectionGroup.session_id == session_id)
        .where(MurfeyDB.DataCollectionGroup.tag == dcg_name)
    ).one()

    clem_img_series: MurfeyDB.CLEMImageSeries = get_db_entry(
        db=murfey_db,
        table=MurfeyDB.CLEMImageSeries,
        session_id=session_id,
        series_name=result.series_name,
    )
    clem_img_series.dcg_id = dcg_entry.id
    clem_img_series.dcg_name = dcg_entry.tag
    murfey_db.add(clem_img_series)
    murfey_db.commit()
    murfey_db.close()


def _register_grid_square(
    session_id: int,
    result: CLEMPreprocessingResult,
    murfey_db: Session,
):
    # Skip this step if no transport manager object is configured
    if _transport_object is None:
        logger.error("Unable to find transport manager")
        return
    # Load all entries for the current data collection group
    dcg_name = result.series_name.split("--")[0]
    if result.series_name.split("--")[1].isdigit():
        dcg_name += f"--{result.series_name.split('--')[1]}"

    # Check if an atlas has been registered
    if atlas_search := murfey_db.exec(
        select(MurfeyDB.CLEMImageSeries)
        .where(MurfeyDB.CLEMImageSeries.session_id == session_id)
        .where(MurfeyDB.CLEMImageSeries.dcg_name == dcg_name)
        .where(MurfeyDB.CLEMImageSeries.data_type == "atlas")
    ).all():
        atlas_entry = atlas_search[0]
    else:
        logger.info(
            f"No atlas has been registered for data collection group {dcg_name!r} yet"
        )
        return

    # Check if there are CLEM entries to register
    if clem_img_series_to_register := murfey_db.exec(
        select(MurfeyDB.CLEMImageSeries)
        .where(MurfeyDB.CLEMImageSeries.session_id == session_id)
        .where(MurfeyDB.CLEMImageSeries.dcg_name == dcg_name)
        .where(MurfeyDB.CLEMImageSeries.data_type == "grid_square")
    ):
        if (
            atlas_entry.x0 is not None
            and atlas_entry.x1 is not None
            and atlas_entry.y0 is not None
            and atlas_entry.y1 is not None
            and atlas_entry.thumbnail_pixels_x is not None
            and atlas_entry.thumbnail_pixels_y is not None
        ):
            atlas_width_real = atlas_entry.x1 - atlas_entry.x0
            atlas_height_real = atlas_entry.y1 - atlas_entry.y0
        else:
            logger.warning("Atlas entry not populated with required values")
            return

        for clem_img_series in clem_img_series_to_register:
            # Register datasets using thumbnail sizes and scales
            if (
                clem_img_series.x0 is not None
                and clem_img_series.x1 is not None
                and clem_img_series.y0 is not None
                and clem_img_series.y1 is not None
            ):
                # Find pixel corresponding to image midpoint on atlas
                x_mid_real = (
                    0.5 * (clem_img_series.x0 + clem_img_series.x1) - atlas_entry.x0
                )
                x_mid_px = int(
                    x_mid_real / atlas_width_real * atlas_entry.thumbnail_pixels_x
                )
                y_mid_real = (
                    0.5 * (clem_img_series.y0 + clem_img_series.y1) - atlas_entry.y0
                )
                y_mid_px = int(
                    y_mid_real / atlas_height_real * atlas_entry.thumbnail_pixels_y
                )

                # Find the size of the image, in pixels, when overlaid on the atlas
                width_scaled = int(
                    (clem_img_series.x1 - clem_img_series.x0)
                    / atlas_width_real
                    * atlas_entry.thumbnail_pixels_x
                )
                height_scaled = int(
                    (clem_img_series.y1 - clem_img_series.y0)
                    / atlas_height_real
                    * atlas_entry.thumbnail_pixels_y
                )
            else:
                logger.warning(
                    f"Image series {clem_img_series.series_name!r} not populated with required values"
                )
                continue

            # Populate grid square Pydantic model
            grid_square_params = GridSquareParameters(
                tag=dcg_name,
                x_location=clem_img_series.x0,
                x_location_scaled=x_mid_px,
                y_location=clem_img_series.y0,
                y_location_scaled=y_mid_px,
                readout_area_x=clem_img_series.image_pixels_x,
                readout_area_y=clem_img_series.image_pixels_y,
                thumbnail_size_x=clem_img_series.thumbnail_pixels_x,
                thumbnail_size_y=clem_img_series.thumbnail_pixels_y,
                width=clem_img_series.image_pixels_x,
                width_scaled=width_scaled,
                height=clem_img_series.image_pixels_y,
                height_scaled=height_scaled,
                x_stage_position=0.5 * (clem_img_series.x0 + clem_img_series.x1),
                y_stage_position=0.5 * (clem_img_series.y0 + clem_img_series.y1),
                pixel_size=clem_img_series.image_pixel_size,
                image=clem_img_series.thumbnail_search_string,
                collection_mode=clem_img_series.collection_mode,
            )
            # Construct colour flags for ISPyB
            color_flags = {
                ispyb_color_flags: int(getattr(clem_img_series, murfey_color_flags, 0))
                for murfey_color_flags, ispyb_color_flags in COLOR_FLAGS_MURFEY_TO_ISPYB.items()
            }
            # Register or update the grid square entry as required
            if grid_square_result := murfey_db.exec(
                select(MurfeyDB.GridSquare)
                .where(MurfeyDB.GridSquare.name == clem_img_series.id)
                .where(MurfeyDB.GridSquare.tag == grid_square_params.tag)
                .where(MurfeyDB.GridSquare.session_id == session_id)
            ).all():
                # Update existing grid square entry on Murfey
                grid_square_entry = grid_square_result[0]
                grid_square_entry.x_location = grid_square_params.x_location
                grid_square_entry.y_location = grid_square_params.y_location
                grid_square_entry.x_stage_position = grid_square_params.x_stage_position
                grid_square_entry.y_stage_position = grid_square_params.y_stage_position
                grid_square_entry.readout_area_x = grid_square_params.readout_area_x
                grid_square_entry.readout_area_y = grid_square_params.readout_area_y
                grid_square_entry.thumbnail_size_x = grid_square_params.thumbnail_size_x
                grid_square_entry.thumbnail_size_y = grid_square_params.thumbnail_size_y
                grid_square_entry.pixel_size = grid_square_params.pixel_size
                grid_square_entry.image = grid_square_params.image

                # Update existing entry on ISPyB
                _transport_object.do_update_grid_square(
                    grid_square_id=grid_square_entry.id,
                    grid_square_parameters=grid_square_params,
                    color_flags=color_flags,
                )
            else:
                # Look up data collection group for current series
                dcg_entry = murfey_db.exec(
                    select(MurfeyDB.DataCollectionGroup)
                    .where(MurfeyDB.DataCollectionGroup.session_id == session_id)
                    .where(MurfeyDB.DataCollectionGroup.tag == grid_square_params.tag)
                ).one()
                # Register to ISPyB
                grid_square_ispyb_result = _transport_object.do_insert_grid_square(
                    atlas_id=dcg_entry.atlas_id,
                    grid_square_id=clem_img_series.id,
                    grid_square_parameters=grid_square_params,
                    color_flags=color_flags,
                )
                # Register to Murfey
                grid_square_entry = MurfeyDB.GridSquare(
                    id=grid_square_ispyb_result.get("return_value", None),
                    name=clem_img_series.id,
                    session_id=session_id,
                    tag=grid_square_params.tag,
                    x_location=grid_square_params.x_location,
                    y_location=grid_square_params.y_location,
                    x_stage_position=grid_square_params.x_stage_position,
                    y_stage_position=grid_square_params.y_stage_position,
                    readout_area_x=grid_square_params.readout_area_x,
                    readout_area_y=grid_square_params.readout_area_y,
                    thumbnail_size_x=grid_square_params.thumbnail_size_x,
                    thumbnail_size_y=grid_square_params.thumbnail_size_y,
                    pixel_size=grid_square_params.pixel_size,
                    image=grid_square_params.image,
                )
            murfey_db.add(grid_square_entry)
            murfey_db.commit()

            # Add grid square ID to existing CLEM image series entry
            clem_img_series.grid_square_id = grid_square_entry.id
            murfey_db.add(clem_img_series)
            murfey_db.commit()
    else:
        logger.info(
            f"No grid squares to register for data collection group {dcg_name!r} yet"
        )
    murfey_db.close()
    return


def run(message: dict, murfey_db: Session) -> dict[str, bool]:
    session_id: int = (
        int(message["session_id"])
        if not isinstance(message["session_id"], int)
        else message["session_id"]
    )
    try:
        if isinstance(message["result"], str):
            json_obj: dict = json.loads(message["result"])
            result = CLEMPreprocessingResult(**json_obj)
        elif isinstance(message["result"], dict):
            result = CLEMPreprocessingResult(**message["result"])
        else:
            logger.error(
                f"Invalid type for TIFF preprocessing result: {type(message['result'])}"
            )
            return {"success": False, "requeue": False}
    except Exception:
        logger.error(
            "Exception encountered when parsing TIFF preprocessing result: \n"
            f"{traceback.format_exc()}"
        )
        return {"success": False, "requeue": False}

    # Outer try-finally block for tidying up database-related section of function
    try:
        try:
            # Load current session from database
            murfey_session = murfey_db.exec(
                select(MurfeyDB.Session).where(MurfeyDB.Session.id == session_id)
            ).one()
        except Exception:
            logger.error(
                "Exception encountered when loading Murfey session information: \n",
                f"{traceback.format_exc()}",
            )
            return {"success": False, "requeue": False}
        try:
            # Register items in Murfey database
            _register_clem_image_series(
                session_id=session_id,
                result=result,
                murfey_db=murfey_db,
            )
        except Exception:
            logger.error(
                "Exception encountered when registering CLEM preprocessing result for "
                f"{result.series_name!r}: \n"
                f"{traceback.format_exc()}"
            )
            return {"success": False, "requeue": False}
        try:
            # Register data collection group and atlas in ISPyB
            _register_dcg_and_atlas(
                session_id=session_id,
                instrument_name=murfey_session.instrument_name,
                visit_name=murfey_session.visit,
                result=result,
                murfey_db=murfey_db,
            )
        except Exception:
            # Log error but allow workflow to proceed
            logger.error(
                "Exception encountered when registering data collection group for CLEM workflow "
                f"using {result.series_name!r}: \n"
                f"{traceback.format_exc()}"
            )

        try:
            # Register CLEM image series as grid squares
            _register_grid_square(
                session_id=session_id,
                result=result,
                murfey_db=murfey_db,
            )
        except Exception:
            # Log error but allow workflow to proceed
            logger.error(
                f"Exception encountered when registering grid square for {result.series_name}: \n"
                f"{traceback.format_exc()}"
            )

        # Construct list of files to use for image alignment and merging steps
        image_combos_to_process = [
            list(result.output_files.values())  # Composite image of all channels
        ]
        if ("gray" in result.output_files.keys()) and len(result.output_files) > 1:
            # Create additional fluorescent-only composite image
            image_combos_to_process.append(
                [
                    file
                    for channel, file in result.output_files.items()
                    if channel != "gray"
                ]
            )
            # Create additional bright field-only image
            image_combos_to_process.append(
                [
                    file
                    for channel, file in result.output_files.items()
                    if channel == "gray"
                ]
            )

        # Request for image alignment and processing for the requested combinations
        for image_combo in image_combos_to_process:
            try:
                submit_cluster_request(
                    session_id=session_id,
                    instrument_name=murfey_session.instrument_name,
                    series_name=result.series_name,
                    images=image_combo,
                    metadata=result.metadata,
                    crop_to_n_frames=processing_params.crop_to_n_frames,
                    align_self=processing_params.align_self,
                    flatten=processing_params.flatten,
                    align_across=processing_params.align_across,
                    messenger=_transport_object,
                )
            except Exception:
                logger.error(
                    "Error requesting image alignment and merging job for "
                    f"{result.series_name!r} series",
                    exc_info=True,
                )
                return {"success": False, "requeue": False}
        logger.info(
            "Successfully requested image alignment and merging job for "
            f"{result.series_name!r} series"
        )
        return {"success": True}

    finally:
        murfey_db.close()



import logging
from importlib.metadata import entry_points
from pathlib import Path

from sqlmodel import select
from sqlmodel.orm.session import Session as SQLModelSession

from murfey.server import _transport_object
from murfey.util.db import DataCollectionGroup

logger = logging.getLogger("murfey.workflows.register_atlas_update")


def run(
    message: dict,
    murfey_db: SQLModelSession,  # Defined for compatibility but unused
):
    if _transport_object is None:
        logger.error("Unable to find transport manager")
        return {"success": False, "requeue": False}

    logger.info(f"Registering updated atlas: \n{message}")

    _transport_object.do_update_atlas(
        atlas_id=message["atlas_id"],
        atlas_image=message["atlas"],
        pixel_size=message["atlas_pixel_size"],
        slot=message["sample"],
        # Extract optional parameters
        collection_mode=message.get("collection_mode"),
        color_flags=message.get("color_flags", {}),
    )

    # Find out how many dcgs we have with this atlas
    if (
        message.get("atlas")
        and message.get("sample")
        and "atlas" in Path(message.get("tag", "/")).parts
    ):
        dcgs_atlas = murfey_db.exec(
            select(DataCollectionGroup)
            .where(DataCollectionGroup.session_id == message["session_id"])
            .where(DataCollectionGroup.atlas == message["atlas"])
            .where(DataCollectionGroup.sample == message["sample"])
        ).all()
        if len(dcgs_atlas) > 1:
            # Skip hooks if this is an atlas and there is a processing dcg present
            logger.info(f"Skipping data collection group hooks for {message['tag']}")
            return {"success": True}

    if dcg_hooks := entry_points(group="murfey.hooks", name="data_collection_group"):
        try:
            for hook in dcg_hooks:
                hook.load()(message["dcgid"], session_id=message["session_id"])
        except Exception:
            logger.error("Call to data collection group hook failed", exc_info=True)
    return {"success": True}



from unittest import mock

from ispyb.sqlalchemy import BLSession, DataCollectionGroup, Proposal
from pytest import mark
from sqlalchemy import select
from sqlalchemy.orm import Session

from murfey.server.ispyb import TransportManager, get_proposal_id, get_session_id
from tests.conftest import ExampleVisit, get_or_create_db_entry


def test_get_session_id(
    ispyb_db_session: Session,
):
    # Manually get the BLSession ID for comparison
    bl_session_id = (
        ispyb_db_session.execute(
            select(BLSession)
            .join(Proposal)
            .where(BLSession.proposalId == Proposal.proposalId)
            .where(BLSession.beamLineName == ExampleVisit.instrument_name)
            .where(Proposal.proposalCode == ExampleVisit.proposal_code)
            .where(Proposal.proposalNumber == str(ExampleVisit.proposal_number))
            .where(BLSession.visit_number == ExampleVisit.visit_number)
        )
        .scalar_one()
        .sessionId
    )

    # Test function
    result = get_session_id(
        microscope=ExampleVisit.instrument_name,
        proposal_code=ExampleVisit.proposal_code,
        proposal_number=str(ExampleVisit.proposal_number),
        visit_number=str(ExampleVisit.visit_number),
        db=ispyb_db_session,
    )
    assert bl_session_id == result


def test_get_proposal_id(
    ispyb_db_session: Session,
):
    # Manually query the Proposal ID
    proposal_id = (
        ispyb_db_session.execute(
            select(Proposal)
            .where(Proposal.proposalCode == ExampleVisit.proposal_code)
            .where(Proposal.proposalNumber == str(ExampleVisit.proposal_number))
        )
        .scalar_one()
        .proposalId
    )

    # Test function
    result = get_proposal_id(
        proposal_code=ExampleVisit.proposal_code,
        proposal_number=str(ExampleVisit.proposal_number),
        db=ispyb_db_session,
    )
    assert proposal_id == result


@mark.skip
def test_get_sub_samples_from_visit():
    pass


@mark.skip
def test_get_all_ongoing_visits():
    pass


@mock.patch("workflows.transport.pika_transport.PikaTransport")
def test_update_data_collection_group(mock_transport, ispyb_db_session: Session):
    # Manually get the BLSession ID for comparison
    bl_session_id = (
        ispyb_db_session.execute(
            select(BLSession)
            .join(Proposal)
            .where(BLSession.proposalId == Proposal.proposalId)
            .where(BLSession.beamLineName == ExampleVisit.instrument_name)
            .where(Proposal.proposalCode == ExampleVisit.proposal_code)
            .where(Proposal.proposalNumber == str(ExampleVisit.proposal_number))
            .where(BLSession.visit_number == ExampleVisit.visit_number)
        )
        .scalar_one()
        .sessionId
    )
    # Insert data collection group
    get_or_create_db_entry(
        session=ispyb_db_session,
        table=DataCollectionGroup,
        insert_kwargs={
            "dataCollectionGroupId": 1,
            "sessionId": bl_session_id,
            "experimentTypeId": 1,
        },
    )

    transport_manager = TransportManager("PikaTransport")
    with mock.patch("murfey.server.ispyb.ISPyBSession", return_value=ispyb_db_session):
        transport_manager.do_update_data_collection_group(
            record=DataCollectionGroup(dataCollectionGroupId=1, experimentTypeId=2)
        )

    final_dcg_entry = get_or_create_db_entry(
        session=ispyb_db_session,
        table=DataCollectionGroup,
        lookup_kwargs={"dataCollectionGroupId": 1},
    )
    assert final_dcg_entry.experimentTypeId == 2



"""
CLI to generate a manifest of the FastAPI router paths present in both the instrument
server and backend server to enable lookup of the URLs based on function name.
"""

import contextlib
import importlib
import inspect
import io
import pkgutil
from argparse import ArgumentParser
from pathlib import Path
from types import ModuleType
from typing import Annotated, Union, get_args, get_origin

import yaml
from fastapi import APIRouter

import murfey
from murfey.cli import PrettierDumper


def extract_base_type(annotation):
    """
    Given a Python type annotation, return its underlying base type.

    This function unwraps `typing.Annotated` to extract the annotated type
    and simplifies `Optional[T]` / `Union[T, None]` to `T`. All other union
    types and complex annotations are returned unchanged.

    Parameters
    ----------
    annotation:
        A Python type annotation (e.g. int, Annotated[int, ...], Optional[int],
        Union[int, str])

    Returns
    -------
    The unwrapped base type, or the original annotation if no unambiguous base
    type can be determined.
    """
    # Unwrap Annotated type annotations
    if get_origin(annotation) is Annotated:
        annotation = get_args(annotation)[0]

    # Unwrap and return single-type Optional type annotations
    if get_origin(annotation) is Union:
        args = [a for a in get_args(annotation) if a is not type(None)]
        if len(args) == 1:
            return args[0]

    # Return complex multi-type annotations or simple unpacked ones
    return annotation


def find_routers(name: str) -> dict[str, APIRouter]:
    def _extract_routers_from_module(module: ModuleType):
        routers = {}
        for name, obj in inspect.getmembers(module):
            if isinstance(obj, APIRouter):
                module_path = module.__name__
                key = f"{module_path}.{name}"
                routers[key] = obj
        return routers

    routers = {}

    # Silence output during import and only return messages if imports fail
    buffer = io.StringIO()
    with contextlib.redirect_stdout(buffer), contextlib.redirect_stderr(buffer):
        # Import the module or package
        try:
            root = importlib.import_module(name)
        except Exception as e:
            captured_logs = buffer.getvalue().strip()
            message = f"Cannot import '{name}': {e}"
            if captured_logs:
                message += f"\n--- Captured output ---\n{captured_logs}"
            raise ImportError(message) from e

        # If it's a package, walk through submodules and extract routers from each
        if hasattr(root, "__path__"):
            module_list = pkgutil.walk_packages(root.__path__, prefix=name + ".")
            for _, module_name, _ in module_list:
                try:
                    module = importlib.import_module(module_name)
                except Exception as e:
                    captured_logs = buffer.getvalue().strip()
                    message = f"Cannot import '{name}': {e}"
                    if captured_logs:
                        message += f"\n--- Captured output ---\n{captured_logs}"
                    raise ImportError(message) from e
                routers.update(_extract_routers_from_module(module))

        # Extract directly from single module
        else:
            routers.update(_extract_routers_from_module(root))

    return routers


def get_route_manifest(routers: dict[str, APIRouter]):
    manifest = {}

    for router_name, router in routers.items():
        routes = []
        for route in router.routes:
            path_params = []
            for param in route.dependant.path_params:
                param_type = extract_base_type(param._type_adapter._type)
                param_info = {
                    "name": param.name if hasattr(param, "name") else "",
                    "type": (
                        param_type.__name__
                        if hasattr(param_type, "__name__")
                        else str(param_type)
                    ),
                }
                path_params.append(param_info)
            for route_dependency in route.dependant.dependencies:
                for param in route_dependency.path_params:
                    param_type = extract_base_type(param._type_adapter._type)
                    param_info = {
                        "name": param.name if hasattr(param, "name") else "",
                        "type": (
                            param_type.__name__
                            if hasattr(param_type, "__name__")
                            else str(param_type)
                        ),
                    }
                    path_params.append(param_info)
            route_info = {
                "path": route.path if hasattr(route, "path") else "",
                "function": route.name if hasattr(route, "name") else "",
                "path_params": path_params,
                "methods": list(route.methods) if hasattr(route, "methods") else [],
            }
            routes.append(route_info)
        manifest[router_name] = routes
    return manifest


def run():
    # Set up additional args
    parser = ArgumentParser()
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help=("Outputs the modules being inspected when creating the route manifest"),
    )
    args = parser.parse_args()

    # Find routers
    print("Finding routers...")
    routers = {
        **find_routers("murfey.instrument_server.api"),
        **find_routers("murfey.server.api"),
    }
    # Generate the manifest
    print("Extracting route information")
    manifest = get_route_manifest(routers)

    # Verify
    if args.debug:
        for router_name, routes in manifest.items():
            print(f"Routes found in {router_name!r}")
            for route in routes:
                for key, value in route.items():
                    print(f"\t{key}: {value}")
                print()

    # Save the manifest
    murfey_dir = Path(murfey.__path__[0])
    manifest_file = murfey_dir / "util" / "route_manifest.yaml"
    with open(manifest_file, "w") as file:
        yaml.dump(
            manifest,
            file,
            Dumper=PrettierDumper,
            default_flow_style=False,
            sort_keys=False,
            indent=2,
        )
    print(
        "Route manifest for instrument and backend servers saved to "
        f"{str(manifest_file)!r}"
    )
    exit()


if __name__ == "__main__":
    run()



import logging
import time

import ispyb.sqlalchemy._auto_db_schema as ISPyBDB
from sqlmodel import select
from sqlmodel.orm.session import Session as SQLModelSession

import murfey.util.db as MurfeyDB
from murfey.server import _transport_object
from murfey.server.ispyb import ISPyBSession, get_session_id
from murfey.util import sanitise

logger = logging.getLogger("murfey.workflows.register_data_collection")


def run(message: dict, murfey_db: SQLModelSession) -> dict[str, bool]:
    # Fail immediately if transport manager was not provided
    if _transport_object is None:
        logger.error("Unable to find transport manager")
        return {"success": False, "requeue": False}

    logger.info(f"Registering the following data collection: \n{message}")

    murfey_session_id = message["session_id"]
    ispyb_session_id = get_session_id(
        microscope=message["microscope"],
        proposal_code=message["proposal_code"],
        proposal_number=message["proposal_number"],
        visit_number=message["visit_number"],
        db=ISPyBSession(),
    )
    dcg = murfey_db.exec(
        select(MurfeyDB.DataCollectionGroup)
        .where(MurfeyDB.DataCollectionGroup.session_id == murfey_session_id)
        .where(MurfeyDB.DataCollectionGroup.tag == message["source"])
    ).all()
    if dcg:
        dcgid = dcg[0].id
        # flush_data_collections(message["source"], murfey_db)
    else:
        time.sleep(2)
        logger.warning(
            "No data collection group ID was found for image directory "
            f"{sanitise(message['image_directory'])} and source "
            f"{sanitise(message['source'])}"
        )
        return {"success": False, "requeue": True}

    if dc_murfey := murfey_db.exec(
        select(MurfeyDB.DataCollection)
        .where(MurfeyDB.DataCollection.tag == message.get("tag"))
        .where(MurfeyDB.DataCollection.dcg_id == dcgid)
    ).all():
        dcid = dc_murfey[0].id
    else:
        if ispyb_session_id is None:
            murfey_dc = MurfeyDB.DataCollection(
                tag=message.get("tag"),
                dcg_id=dcgid,
            )
        else:
            record = ISPyBDB.DataCollection(
                SESSIONID=ispyb_session_id,
                experimenttype=message["experiment_type"],
                imageDirectory=message["image_directory"],
                imageSuffix=message["image_suffix"],
                voltage=message["voltage"],
                dataCollectionGroupId=dcgid,
                pixelSizeOnImage=message["pixel_size"],
                imageSizeX=message["image_size_x"],
                imageSizeY=message["image_size_y"],
                slitGapHorizontal=message.get("slit_width"),
                magnification=message.get("magnification"),
                exposureTime=message.get("exposure_time"),
                totalExposedDose=message.get("total_exposed_dose"),
                c2aperture=message.get("c2aperture"),
                phasePlate=int(message.get("phase_plate", 0)),
            )
            dcid = _transport_object.do_insert_data_collection(
                record,
                tag=(
                    message.get("tag")
                    if message["experiment_type"] == "tomography"
                    else ""
                ),
            ).get("return_value", None)
            if dcid is None:
                time.sleep(2)
                logger.error(
                    "Failed to register the following data collection: \n"
                    f"{message} \n"
                    "Requeueing message"
                )
                return {"success": False, "requeue": True}
            murfey_dc = MurfeyDB.DataCollection(
                id=dcid,
                tag=message.get("tag"),
                dcg_id=dcgid,
            )
        murfey_db.add(murfey_dc)
        murfey_db.commit()
        murfey_db.close()
    return {"success": True}



import os
import queue
import threading
from pathlib import Path

import pytest

from murfey.client.watchdir import DirWatcher
from tests.conftest import ExampleVisit


def test_dirwatcher_initialises(tmp_path: Path):
    # Check that the DirWatcher initialises with the default attributes
    watcher = DirWatcher(path=str(tmp_path))
    assert watcher._basepath == os.fspath(str(tmp_path))
    assert watcher._lastscan == {}
    assert watcher._file_candidates == {}
    assert watcher.settling_time == 60
    assert watcher._appearance_time is None
    assert watcher._substrings_blacklist == {}
    assert watcher._transfer_all is True
    assert watcher._modification_overwrite is None
    assert isinstance(watcher._init_time, float)
    assert isinstance(watcher.queue, queue.Queue)
    assert isinstance(watcher.thread, threading.Thread)
    assert watcher._stopping is False
    assert watcher._halt_thread is False

    # Check that the string representation is as expected
    assert str(watcher) == f"<DirWatcher ({os.fspath(str(tmp_path))})>"


@pytest.fixture
def clem_visit_dir(tmp_path: Path):
    visit_name = f"{ExampleVisit.proposal_code}{ExampleVisit.proposal_number}-{ExampleVisit.visit_number}"
    visit_dir = tmp_path / "clem" / "data" / "2025" / visit_name
    visit_dir.mkdir(parents=True, exist_ok=True)
    return visit_dir


@pytest.fixture
def clem_test_files(clem_visit_dir: Path):
    # Create test files for the DirWatcher to scan
    file_list: list[Path] = []
    project_dir = clem_visit_dir / "images" / "test_grid"

    # Example atlas collection
    for s in range(20):
        file_list.append(
            project_dir
            / "Overview 1"
            / "Image 1"
            / f"Image 1--Stage{str(s).zfill(2)}.tif"
        )
    file_list.append(
        project_dir / "Overview 1" / "Image 1" / "Metadata" / "Image 1.xlif"
    )

    # Example image stack collection
    for c in range(3):
        for z in range(10):
            file_list.append(
                project_dir
                / "TileScan 1"
                / "Position 1"
                / f"Position 1--C{str(c).zfill(2)}--Z{str(z).zfill(2)}.tif"
            )
    file_list.append(
        project_dir / "TileScan 1" / "Position 1" / "Metadata" / "Position 1.xlif"
    )

    # Create all files and directories specified
    for file in file_list:
        if not file.parent.exists():
            file.parent.mkdir(parents=True)
        if not file.exists():
            file.touch()
    return sorted(file_list)


@pytest.fixture
def clem_junk_files(clem_visit_dir: Path):
    # Create junk files that are to be blacklisted from the CLEM workflow
    file_list: list[Path] = []
    project_dir = clem_visit_dir / "images" / "test_grid"

    # Create junk atlas data
    for n in range(5):
        for s in range(20):
            file_list.append(
                project_dir
                / "Image 1"
                / f"Image 1_pmd_{n}"
                / f"Image 1_pmd_{n}--Stage{str(s).zfill(2)}.tif"
            )
        file_list.append(
            project_dir / "Image 1" / f"Image 1_pmd_{n}" / "Metadata" / "Image 1.xlif"
        )

    # Creat junk image data
    for n in range(5):
        for c in range(3):
            for z in range(10):
                file_list.append(
                    project_dir
                    / "Position 1"
                    / f"Position 1_pmd_{n}"
                    / f"Position 1_pmd_{n}--C{str(c).zfill(2)}--Z{str(z).zfill(2)}.tif"
                )
        file_list.append(
            project_dir
            / "Position 1"
            / f"Position 1_pmd_{n}"
            / "Metadata"
            / "Position 1.xlif"
        )

    # Create remaining junk files
    for file_path in (
        "1.xlef",
        "Metadata/IOManagerConfiguation.xlif",
        "Metadata/Overview 1.xlcf",
        "Metadata/TileScan 1.xlcf",
        "Overview 1/Image 1/Image 1_histo.lof",
        "TileScan 1/Position 1/Position 1_histo.lof",
        "Overview 1/Image 1/Metadata/Image 1_histo.xlif",
        "TileScan 1/Position 1/Metadata/Position 1_histo.xlif",
    ):
        file_list.append(project_dir / file_path)

    # Create files and directoriees
    for file in file_list:
        if not file.parent.exists():
            file.parent.mkdir(parents=True)
        if not file.exists():
            file.touch()
    return sorted(file_list)


scan_directory_params_matrix: tuple[tuple[str, dict[str, list[str]]], ...] = (
    # Workflow type | Substrings blacklist
    (
        "clem",
        {
            "directories": [
                "_pmd_",
            ],
            "files": [
                ".xlef",
                ".xlcf",
                "_histo.lof",
                "_histo.xlif",
                "IOManagerConfiguation.xlif",
            ],
        },
    ),
)


@pytest.mark.parametrize("test_params", scan_directory_params_matrix)
def test_scan_directory(
    clem_visit_dir: Path,
    clem_test_files: list[Path],
    clem_junk_files: list[Path],
    test_params: tuple[str, dict[str, list[str]]],
):
    # Unpack test params
    workflow_type, substrings_blacklist = test_params

    # Initialise different watchers based on the workflow to test and run the scan
    if workflow_type == "clem":
        watcher = DirWatcher(
            path=str(clem_visit_dir),
            substrings_blacklist=substrings_blacklist,
        )
        result = watcher._scan_directory()

        # Check that the result does not contain the junk files
        assert [str(file) for file in clem_test_files] == sorted(result.keys())



from pathlib import Path
from unittest import mock

import pytest
from pytest_mock import MockerFixture

from murfey.client.destinations import (
    determine_default_destination,
    find_longest_data_directory,
)


@pytest.mark.parametrize(
    "test_params",
    (
        # File to match | Use Path? | Expected result
        (
            "X:/cm12345-6/Supervisor/Images-Disc1/file",
            True,
            ("X:/", "cm12345-6/Supervisor/Images-Disc1"),
        ),
        (
            "X:/DATA/cm12345-6/Supervisor/Images-Disc1/file",
            False,
            ("X:/DATA", "cm12345-6/Supervisor/Images-Disc1"),
        ),
        (
            "X:/DoseFractions/cm12345-6/Supervisor/Images-Disc1/file",
            True,
            ("X:/DoseFractions", "cm12345-6/Supervisor/Images-Disc1"),
        ),
        (
            "X:/DoseFractions/DATA/cm12345-6/Supervisor/Images-Disc1/file",
            False,
            ("X:/DoseFractions/DATA", "cm12345-6/Supervisor/Images-Disc1"),
        ),
    ),
)
def test_find_longest_data_directory(
    mocker: MockerFixture, test_params: tuple[str, bool, tuple[str, str]]
):
    # Unpack test params
    match_path, use_path, (expected_base_dir, expected_mid_dir) = test_params

    # Construct data directories using strings or Paths as needed
    data_directories: list[str] | list[Path] = [
        "X:",
        "X:/DATA",
        "X:/DoseFractions",
        "X:/DoseFractions/DATA",
    ]
    if use_path:
        data_directories = [Path(dd) for dd in data_directories]
    # Patch Pathlib's 'absolute()' function, since we are simulating Windows on Linux
    mocker.patch("murfey.client.destinations.Path.absolute", lambda self: self)

    base_dir, mid_dir = find_longest_data_directory(
        Path(match_path),
        data_directories,
    )
    assert base_dir == Path(expected_base_dir)
    assert mid_dir == Path(expected_mid_dir)


source_list = [
    ["X:/DoseFractions/cm12345-6/Supervisor", "Supervisor", True, "extra_name"],
    ["X:/DoseFractions/Supervisor/Images-Disc1", "Supervisor", False, ""],
    [
        "X:/DoseFractions/DATA/Supervisor/Sample1",
        "Supervisor_Sample1",
        False,
        "extra_name",
    ],
]


@mock.patch("murfey.client.destinations.capture_get")
@mock.patch("murfey.client.destinations.capture_post")
@pytest.mark.parametrize("sources", source_list)
def test_determine_default_destinations_suggested_path(mock_post, mock_get, sources):
    mock_environment = mock.Mock()
    mock_environment.murfey_session = 2
    mock_environment.instrument_name = "m01"
    mock_environment.destination_registry = {}

    source, source_name, touch, extra_directory = sources

    mock_get().json.return_value = {
        "data_directories": ["X:/DoseFractions", "X:/DoseFractions/DATA"]
    }
    mock_post().json.return_value = {"suggested_path": "/base_path/2025/cm12345-6/raw"}

    destination = determine_default_destination(
        visit="cm12345-6",
        source=Path(source),
        destination="2025",
        environment=mock_environment,
        token="token",
        touch=touch,
        extra_directory=extra_directory,
        use_suggested_path=True,
    )
    mock_get.assert_any_call(
        base_url=str(mock_environment.url.geturl()),
        router_name="session_control.router",
        function_name="machine_info_by_instrument",
        token="token",
        instrument_name="m01",
    )
    mock_post.assert_any_call(
        base_url=str(mock_environment.url.geturl()),
        router_name="file_io_instrument.router",
        function_name="suggest_path",
        token="token",
        visit_name="cm12345-6",
        session_id=2,
        data={
            "base_path": "2025/cm12345-6/raw",
            "touch": touch,
            "extra_directory": extra_directory,
        },
    )

    assert destination == f"/base_path/2025/cm12345-6/raw/{extra_directory}"
    assert (
        mock_environment.destination_registry.get(source_name)
        == "/base_path/2025/cm12345-6/raw"
    )


@mock.patch("murfey.client.destinations.capture_get")
@mock.patch("murfey.client.destinations.capture_post")
def test_determine_default_destinations_short_path(mock_post, mock_get):
    """Test for the case where the data directory is a drive"""
    mock_environment = mock.Mock()
    mock_environment.murfey_session = 2
    mock_environment.instrument_name = "m01"
    mock_environment.destination_registry = {}

    mock_get().json.return_value = {"data_directories": ["X:/"]}
    mock_post().json.return_value = {"suggested_path": "/base_path/2025/cm12345-6/raw"}

    destination = determine_default_destination(
        visit="cm12345-6",
        source=Path("X:/cm12345-6/Supervisor"),
        destination="2025",
        environment=mock_environment,
        token="token",
        touch=True,
        extra_directory="",
        use_suggested_path=True,
    )
    mock_get.assert_any_call(
        base_url=str(mock_environment.url.geturl()),
        router_name="session_control.router",
        function_name="machine_info_by_instrument",
        token="token",
        instrument_name="m01",
    )
    mock_post.assert_any_call(
        base_url=str(mock_environment.url.geturl()),
        router_name="file_io_instrument.router",
        function_name="suggest_path",
        token="token",
        visit_name="cm12345-6",
        session_id=2,
        data={
            "base_path": "2025/cm12345-6/raw",
            "touch": True,
            "extra_directory": "",
        },
    )

    assert destination == "/base_path/2025/cm12345-6/raw/"
    assert (
        mock_environment.destination_registry.get("Supervisor")
        == "/base_path/2025/cm12345-6/raw"
    )


@mock.patch("murfey.client.destinations.capture_get")
@pytest.mark.parametrize("sources", source_list)
def test_determine_default_destinations_skip_suggested(mock_get, sources):
    mock_environment = mock.Mock()
    mock_environment.murfey_session = 2
    mock_environment.instrument_name = "m01"
    mock_environment.destination_registry = {}

    source, source_name, touch, extra_directory = sources

    mock_get().json.return_value = {
        "data_directories": ["X:/DoseFractions", "X:/DoseFractions/DATA"]
    }

    destination = determine_default_destination(
        visit="cm12345-6",
        source=Path(source),
        destination="2025",
        environment=mock_environment,
        token="token",
        touch=touch,
        extra_directory=extra_directory,
        use_suggested_path=False,
    )
    mock_get.assert_any_call(
        base_url=str(mock_environment.url.geturl()),
        router_name="session_control.router",
        function_name="machine_info_by_instrument",
        token="token",
        instrument_name="m01",
    )

    assert destination == f"2025/cm12345-6/{Path(source).name}/{extra_directory}"


parameter_list_fail_cases = [
    ["X:/DoseFractions/cm12345-6", "", "2025", "X:/DoseFractions"],
    ["X:/DoseFractions/cm12345-6", "cm12345-6", "", "X:/DoseFractions"],
    ["X:/DoseFractions", "cm12345-6", "2025", "X:/DoseFractions"],
    ["X:/cm12345-6", "cm12345-6", "2025", "X:/DoseFractions"],
]


@mock.patch("murfey.client.destinations.capture_get")
@pytest.mark.parametrize("destination_params", parameter_list_fail_cases)
def test_determine_default_destinations_failures(mock_get, destination_params):
    """
    Test failure of the following cases:
    No visit, no destination, source = default, source not in default
    """
    mock_get().json.return_value = {
        "data_directories": ["X:/DoseFractions", "X:/DoseFractions/DATA"]
    }
    source, visit, destination, default_dests = destination_params
    mock_environment = mock.Mock()
    with pytest.raises(ValueError):
        determine_default_destination(
            visit=visit,
            source=Path(source),
            destination=destination,
            environment=mock_environment,
            token="token",
        )



from logging import getLogger
from typing import Optional

import sqlalchemy
from fastapi import APIRouter, Depends
from pydantic import BaseModel
from sqlmodel import Session, select

from murfey.server.api.auth import (
    MurfeySessionIDFrontend as MurfeySessionID,
    validate_token,
)
from murfey.server.murfey_db import murfey_db
from murfey.util.db import SessionProcessingParameters

logger = getLogger("murfey.server.api.processing_parameters")

router = APIRouter(
    prefix="/session_parameters",
    dependencies=[Depends(validate_token)],
    tags=["Processing Parameters"],
)


class EditableSessionProcessingParameters(BaseModel):
    gain_ref: str = ""
    dose_per_frame: Optional[float] = None
    eer_fractionation_file: str = ""
    symmetry: str = ""
    run_class3d: Optional[bool] = None


@router.get("/sessions/{session_id}/session_processing_parameters")
def get_session_processing_parameters(
    session_id: MurfeySessionID, db: Session = murfey_db
) -> Optional[EditableSessionProcessingParameters]:
    try:
        proc_params = db.exec(
            select(SessionProcessingParameters).where(
                SessionProcessingParameters.session_id == session_id
            )
        ).one()
    except sqlalchemy.exc.NoResultFound:
        return None
    return EditableSessionProcessingParameters(
        gain_ref=proc_params.gain_ref,
        dose_per_frame=proc_params.dose_per_frame,
        eer_fractionation_file=proc_params.eer_fractionation_file,
        symmetry=proc_params.symmetry,
        run_class3d=proc_params.run_class3d,
    )


@router.post("/sessions/{session_id}/session_processing_parameters")
def set_session_processing_parameters(
    session_id: MurfeySessionID,
    edited_parameters: EditableSessionProcessingParameters,
    db: Session = murfey_db,
) -> EditableSessionProcessingParameters:
    proc_params = db.exec(
        select(SessionProcessingParameters).where(
            SessionProcessingParameters.session_id == session_id
        )
    ).one()
    proc_params.gain_ref = edited_parameters.gain_ref or proc_params.gain_ref
    proc_params.dose_per_frame = (
        edited_parameters.dose_per_frame or proc_params.dose_per_frame
    )
    proc_params.eer_fractionation_file = (
        edited_parameters.eer_fractionation_file or proc_params.eer_fractionation_file
    )
    proc_params.symmetry = edited_parameters.symmetry or proc_params.symmetry
    if edited_parameters.run_class3d is not None:
        proc_params.run_class3d = edited_parameters.run_class3d
    db.add(proc_params)
    db.commit()
    db.close()
    return edited_parameters



from unittest.mock import patch
from urllib.parse import urlparse

from murfey.client.contexts.atlas import AtlasContext
from murfey.client.instance_environment import MurfeyInstanceEnvironment


def test_atlas_context_initialisation(tmp_path):
    context = AtlasContext("tomo", tmp_path, "token")
    assert context.name == "Atlas"
    assert context._acquisition_software == "tomo"
    assert context._basepath == tmp_path
    assert context._token == "token"


@patch("murfey.client.contexts.atlas.capture_post")
def test_atlas_context_mrc(mock_capture_post, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path / "cm12345-6"],
        default_destinations={
            tmp_path / "cm12345-6": f"{tmp_path}/destination/cm12345-6"
        },
        instrument_name="",
        visit="cm12345-6",
        murfey_session=1,
    )
    context = AtlasContext("tomo", tmp_path, "token")

    atlas_mrc = tmp_path / "cm12345-6/Supervisor_atlas/Sample2/Atlas/Atlas_1.mrc"
    atlas_mrc.parent.mkdir(parents=True)
    atlas_mrc.touch()

    context.post_transfer(
        atlas_mrc,
        environment=env,
    )
    mock_capture_post.assert_called_once_with(
        base_url="http://localhost:8000",
        router_name="session_control.spa_router",
        function_name="make_atlas_jpg",
        token="token",
        session_id=1,
        data={"path": f"{tmp_path}/destination/{atlas_mrc.relative_to(tmp_path)}"},
    )


@patch("murfey.client.contexts.atlas.capture_post")
def test_atlas_context_xml(mock_capture_post, tmp_path):
    env = MurfeyInstanceEnvironment(
        url=urlparse("http://localhost:8000"),
        client_id=0,
        sources=[tmp_path / "cm12345-6"],
        default_destinations={
            tmp_path / "cm12345-6": f"{tmp_path}/destination/cm12345-6"
        },
        instrument_name="",
        visit="cm12345-6",
        murfey_session=1,
    )
    context = AtlasContext("tomo", tmp_path, "token")

    atlas_pixel_size = 4.6
    atlas_xml = tmp_path / "cm12345-6/Supervisor_atlas/Sample2/Atlas/Atlas_1.xml"
    atlas_xml.parent.mkdir(parents=True)
    with open(atlas_xml, "w") as new_xml:
        new_xml.write(
            f"<MicroscopeImage><SpatialScale><pixelSize><x><numericValue>{atlas_pixel_size}"
            "</numericValue></x></pixelSize></SpatialScale></MicroscopeImage>"
        )

    context.post_transfer(
        atlas_xml,
        environment=env,
    )
    dcg_data = {
        "experiment_type_id": 44,  # Atlas
        "tag": str(atlas_xml.parent),
        "atlas": f"{tmp_path}/destination/{atlas_xml.relative_to(tmp_path).with_suffix('.jpg')}",
        "sample": 2,
        "atlas_pixel_size": atlas_pixel_size * 7.8,
    }
    mock_capture_post.assert_called_once_with(
        base_url="http://localhost:8000",
        router_name="workflow.router",
        function_name="register_dc_group",
        token="token",
        visit_name="cm12345-6",
        session_id=1,
        data=dcg_data,
    )



from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from murfey.workflows.register_data_collection import run
from tests.conftest import ExampleVisit

register_data_collection_params_matrix = (
    # ISPyB session ID return value | DCG search result | DC search result | Insert data collection
    (0, 0, 0, 0),
    (0, 0, 0, None),
    (0, 0, None, 0),
    (0, 0, None, None),
    (0, None, 0, 0),
    (0, None, 0, None),
    (0, None, None, 0),
    (0, None, None, None),
    (None, 0, 0, 0),
    (None, 0, 0, None),
    (None, 0, None, 0),
    (None, 0, None, None),
    (None, None, 0, 0),
    (None, None, 0, None),
    (None, None, None, 0),
    (None, None, None, None),
)


@pytest.mark.parametrize("test_params", register_data_collection_params_matrix)
def test_run(
    mocker: MockerFixture,
    test_params: tuple[int | None, int | None, int | None, int | None],
):
    # Unpack test params
    ispyb_session_id, dcg_result, dc_result, insert_data_collection = test_params

    # Set up mock objects
    # 'get_session_id'
    mock_get_session_id = mocker.patch(
        "murfey.workflows.register_data_collection.get_session_id"
    )
    mock_get_session_id.return_value = ispyb_session_id

    # Transport object inserts
    mock_transport_object = mocker.patch(
        "murfey.workflows.register_data_collection._transport_object"
    )
    mock_transport_object.do_insert_data_collection.return_value = {
        "return_value": insert_data_collection
    }

    # Murfey database
    mock_murfey_db = MagicMock()
    mock_dcg = MagicMock()
    mock_dcg.id = dcg_result

    mock_dc = MagicMock()
    mock_dc.id = dc_result
    mock_murfey_db.exec.return_value.all.side_effect = [
        # Sequence of mock database tables
        [mock_dcg] if dcg_result is not None else [],
        [mock_dc] if dc_result is not None else [],
    ]

    # Run the function and check results and calls
    message = {
        "session_id": 0,
        "microscope": "test_instrument",
        "proposal_code": ExampleVisit.proposal_code,
        "proposal_number": ExampleVisit.proposal_number,
        "visit_number": ExampleVisit.visit_number,
        "source": "some_path",
        "image_directory": "some_path",
        "tag": "some_string",
        "experiment_type": "SPA",
        "image_suffix": ".jpg",
        "voltage": 200,
        "pixel_size": 1e-9,
        "image_size_x": 2048,
        "image_size_y": 2048,
        "slit_width": 0.005,
        "magnification": 150000,
        "exposure_time": 30,
        "total_exposed_dose": 30,
        "c2aperture": 5,
        "phase_plate": 1,
    }
    result = run(message=message, murfey_db=mock_murfey_db)
    if dcg_result is None:
        assert result == {"success": False, "requeue": True}
    else:
        if dc_result is None:
            if ispyb_session_id is not None:
                mock_transport_object.do_insert_data_collection.assert_called_once()
                if insert_data_collection is not None:
                    assert result == {"success": True}
                else:
                    assert result == {"success": False, "requeue": True}
            else:
                mock_transport_object.do_insert_data_collection.assert_not_called()
                assert result == {"success": True}
        else:
            assert result == {"success": True}



from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from murfey.workflows.register_experiment_type_update import run

register_experiment_type_update_matrix = (0, 1, None)


@pytest.mark.parametrize("insert_dcg", register_experiment_type_update_matrix)
def test_run(
    mocker: MockerFixture,
    insert_dcg: int | None,
):
    # Mock the transport object functions
    mock_transport_object = mocker.patch(
        "murfey.workflows.register_experiment_type_update._transport_object"
    )
    mock_transport_object.do_update_data_collection_group.return_value = {
        "return_value": insert_dcg,
    }
    mock_ispyb = mocker.patch(
        "murfey.workflows.register_experiment_type_update.ISPyBDB"
    )
    mock_ispyb.DataCollectionGroup.return_value = "ispyb_dcg"

    # Mock the Murfey database
    mock_murfey_db = MagicMock()

    # Run the function and check the results and calls
    message = {
        "dcgid": 1,
        "experiment_type_id": 0,
    }
    result = run(message=message, murfey_db=mock_murfey_db)
    mock_ispyb.DataCollectionGroup.assert_called_once_with(
        dataCollectionGroupId=1, experimentTypeId=0
    )
    mock_transport_object.do_update_data_collection_group.assert_called_once_with(
        "ispyb_dcg"
    )
    if insert_dcg is not None:
        assert result == {"success": True}
    else:
        assert result == {"success": False, "requeue": True}



from logging import getLogger
from typing import Tuple

import numpy as np
from sqlalchemy import func
from sqlmodel import Session, select

from murfey.server import _transport_object
from murfey.server.feedback import _app_id, _murfey_id
from murfey.util.config import get_machine_config
from murfey.util.db import (
    AutoProcProgram,
    ClassificationFeedbackParameters,
    DataCollection,
    ParticleSizes,
    ProcessingJob,
    Session as MurfeySession,
    TomogramPicks,
    TomographyProcessingParameters,
)
from murfey.util.processing_params import default_tomo_parameters

logger = getLogger("murfey.workflows.tomo.feedback")


def _ids_tomo_classification(
    app_id: int, recipe: str, murfey_db: Session
) -> Tuple[int, int]:
    dcg_id = (
        murfey_db.exec(
            select(AutoProcProgram, ProcessingJob, DataCollection)
            .where(AutoProcProgram.id == app_id)
            .where(AutoProcProgram.pj_id == ProcessingJob.id)
            .where(ProcessingJob.dc_id == DataCollection.id)
        )
        .one()[2]
        .dcg_id
    )
    pj_id = (
        murfey_db.exec(
            select(ProcessingJob, DataCollection)
            .where(DataCollection.dcg_id == dcg_id)
            .where(ProcessingJob.dc_id == DataCollection.id)
            .where(ProcessingJob.recipe == recipe)
        )
        .one()[0]
        .id
    )
    return dcg_id, pj_id


def _register_picked_tomogram_use_diameter(message: dict, murfey_db: Session):
    """Received picked particles from the tomogram autopick service"""
    # Add this message to the table of seen messages
    dcg_id, pj_id = _ids_tomo_classification(
        message["program_id"], "em-tomo-class2d", murfey_db
    )

    pick_params = TomogramPicks(
        pj_id=pj_id,
        tomogram=message["tomogram"],
        cbox_3d=message["cbox_3d"],
        particle_count=message["particle_count"],
        tomogram_pixel_size=message["pixel_size"],
    )
    murfey_db.add(pick_params)
    murfey_db.commit()

    picking_db_len = murfey_db.exec(
        select(func.count(ParticleSizes.id)).where(ParticleSizes.pj_id == pj_id)
    ).one()
    if picking_db_len > default_tomo_parameters.batch_size_2d:
        # If there are enough particles to get a diameter
        instrument_name = (
            murfey_db.exec(
                select(MurfeySession).where(MurfeySession.id == message["session_id"])
            )
            .one()
            .instrument_name
        )
        machine_config = get_machine_config(instrument_name=instrument_name)[
            instrument_name
        ]
        tomo_params = murfey_db.exec(
            select(TomographyProcessingParameters).where(
                TomographyProcessingParameters.dcg_id == dcg_id
            )
        ).one()

        particle_diameter = tomo_params.particle_diameter

        feedback_params = murfey_db.exec(
            select(ClassificationFeedbackParameters).where(
                ClassificationFeedbackParameters.pj_id == pj_id
            )
        ).one()
        if not feedback_params.next_job:
            feedback_params.next_job = 9

        if not particle_diameter:
            # If the diameter has not been calculated then find it
            picking_db = murfey_db.exec(
                select(ParticleSizes.particle_size).where(ParticleSizes.pj_id == pj_id)
            ).all()
            particle_diameter = float(np.quantile(list(picking_db), 0.75))
            tomo_params.particle_diameter = particle_diameter
            murfey_db.add(tomo_params)
            murfey_db.commit()

            tomo_pick_db = murfey_db.exec(
                select(TomogramPicks).where(TomogramPicks.pj_id == pj_id)
            ).all()
            for saved_message in tomo_pick_db:
                # Send on all saved messages to extraction
                class_uuids = {
                    str(i + 1): m
                    for i, m in enumerate(
                        _murfey_id(
                            _app_id(pj_id, murfey_db),
                            murfey_db,
                            number=default_tomo_parameters.nr_classes_2d,
                        )
                    )
                }
                class2d_grp_uuid = _murfey_id(_app_id(pj_id, murfey_db), murfey_db)[0]
                zocalo_message: dict = {
                    "parameters": {
                        "tomogram": saved_message.tomogram,
                        "cbox_3d": saved_message.cbox_3d,
                        "pixel_size": saved_message.tomogram_pixel_size,
                        "particle_diameter": particle_diameter,
                        "kv": tomo_params.voltage,
                        "node_creator_queue": machine_config.node_creator_queue,
                        "session_id": message["session_id"],
                        "autoproc_program_id": _app_id(pj_id, murfey_db),
                        "batch_size": default_tomo_parameters.batch_size_2d,
                        "nr_classes": default_tomo_parameters.nr_classes_2d,
                        "picker_id": None,
                        "class2d_grp_uuid": class2d_grp_uuid,
                        "class_uuids": class_uuids,
                        "next_job": feedback_params.next_job,
                    },
                    "recipes": ["em-tomo-class2d"],
                }
                if _transport_object:
                    zocalo_message["parameters"]["feedback_queue"] = (
                        _transport_object.feedback_queue
                    )
                    _transport_object.send(
                        "processing_recipe", zocalo_message, new_connection=True
                    )
                feedback_params.next_job += 2
                murfey_db.delete(saved_message)
        else:
            # If the diameter is known then just send the new message
            particle_diameter = tomo_params.particle_diameter
            class_uuids = {
                str(i + 1): m
                for i, m in enumerate(
                    _murfey_id(
                        _app_id(pj_id, murfey_db),
                        murfey_db,
                        number=default_tomo_parameters.nr_classes_2d,
                    )
                )
            }
            class2d_grp_uuid = _murfey_id(_app_id(pj_id, murfey_db), murfey_db)[0]
            zocalo_message = {
                "parameters": {
                    "tomogram": message["tomogram"],
                    "cbox_3d": message["cbox_3d"],
                    "pixel_size": message["pixel_size"],
                    "particle_diameter": particle_diameter,
                    "kv": tomo_params.voltage,
                    "node_creator_queue": machine_config.node_creator_queue,
                    "session_id": message["session_id"],
                    "autoproc_program_id": _app_id(pj_id, murfey_db),
                    "batch_size": default_tomo_parameters.batch_size_2d,
                    "nr_classes": default_tomo_parameters.nr_classes_2d,
                    "picker_id": None,
                    "class2d_grp_uuid": class2d_grp_uuid,
                    "class_uuids": class_uuids,
                    "next_job": feedback_params.next_job,
                },
                "recipes": ["em-tomo-class2d"],
            }
            if _transport_object:
                zocalo_message["parameters"]["feedback_queue"] = (
                    _transport_object.feedback_queue
                )
                _transport_object.send(
                    "processing_recipe", zocalo_message, new_connection=True
                )
            feedback_params.next_job += 2
        murfey_db.add(feedback_params)
        murfey_db.commit()
    else:
        # If not enough particles then save the new sizes
        particle_list = message.get("particle_diameters")
        assert isinstance(particle_list, list)
        for particle in particle_list:
            new_particle = ParticleSizes(pj_id=pj_id, particle_size=particle)
            murfey_db.add(new_particle)
            murfey_db.commit()
    murfey_db.close()


def picked_tomogram(message: dict, murfey_db: Session) -> dict[str, bool]:
    _register_picked_tomogram_use_diameter(message, murfey_db)
    return {"success": True}



from logging import getLogger
from pathlib import Path

from fastapi import APIRouter, Depends
from pydantic import BaseModel
from sqlmodel import select

from murfey.server.api.auth import (
    MurfeySessionIDFrontend as MurfeySessionID,
    validate_token,
)
from murfey.server.api.file_io_shared import (
    GainReference,
    process_gain as _process_gain,
)
from murfey.server.murfey_db import murfey_db
from murfey.util import secure_path
from murfey.util.config import get_machine_config
from murfey.util.db import Session

logger = getLogger("murfey.server.api.file_io_frontend")


router = APIRouter(
    prefix="/file_io/frontend",
    dependencies=[Depends(validate_token)],
    tags=["File I/O: Frontend"],
)


@router.post("/sessions/{session_id}/process_gain")
async def process_gain(
    session_id: MurfeySessionID, gain_reference_params: GainReference, db=murfey_db
):
    result = await _process_gain(session_id, gain_reference_params, db)
    return result


class SymlinkParameters(BaseModel):
    target: Path  # these are the paths without the rsync basepath as that is what the frontend has access to
    symlink: Path
    override: bool = False


@router.post("/sessions/{session_id}/symlink")
async def create_symlink(
    session_id: MurfeySessionID, symlink_params: SymlinkParameters, db=murfey_db
) -> str:
    murfey_session = db.exec(select(Session).where(Session.id == session_id)).one()
    instrument_name = murfey_session.instrument_name
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    rsync_basepath = (machine_config.rsync_basepath or Path("")).resolve()
    symlink_full_path = secure_path(
        rsync_basepath / symlink_params.symlink, keep_spaces=True
    )
    # Verify that the symlink provided does not lead elsewhere
    if not symlink_full_path.resolve().is_relative_to(rsync_basepath):
        logger.warning(
            "Symlink rejected because it will be created in a forbidden location"
        )
        return ""
    # Remove and replace symlink if it exists are 'override' is set
    if symlink_full_path.is_symlink() and symlink_params.override:
        symlink_full_path.unlink()
    # If a file/folder already exists using the desired symlink name, return empty string
    if symlink_full_path.exists():
        return ""
    symlink_full_path.symlink_to(
        secure_path(rsync_basepath / symlink_params.target, keep_spaces=True)
    )
    return str(symlink_params.symlink)



import copy
import secrets
from unittest.mock import AsyncMock, MagicMock

import pytest
from fastapi import HTTPException
from pytest_mock import MockerFixture
from sqlmodel import Session as SQLModelSession

from murfey.server.api.auth import (
    check_user,
    get_visit_name,
    submit_to_auth_endpoint,
    validate_frontend_session_access,
    validate_session_against_visit,
    validate_token,
    validate_user,
    validate_user_instrument_access,
)
from murfey.util.db import MurfeyUser, Session as MurfeySession


@pytest.mark.parametrize(
    "test_params",
    (  # User to check | Expected result
        ("murfey_user", True),
        ("some_dud", False),
    ),
)
def test_check_user(
    mocker: MockerFixture,
    murfey_db_session: SQLModelSession,
    test_params: tuple[str, bool],
):
    # Unpack test params
    user_to_check, expected_result = test_params

    # Add the test user to the database
    murfey_user = MurfeyUser(
        username="murfey_user",
        hashed_password="asdfghjkl",
    )
    murfey_db_session.add(murfey_user)
    murfey_db_session.commit()

    # Patch the Session context
    mock_session_context = MagicMock()
    mock_session_context.__enter__.return_value = murfey_db_session
    mock_session_context.__exit__.return_value = None
    mocker.patch("murfey.server.api.auth.Session", return_value=mock_session_context)

    # Run the function and check that the result is as expected
    result = check_user(user_to_check)
    assert result == expected_result


@pytest.mark.parametrize(
    "test_params",
    (  # URL subpath | Auth type | Status code | Validation result
        (
            "validate_token",
            "cookie",
            200,
            True,
        ),
        (
            "validate_visit_access/some_visit",
            "password",
            200,
            True,
        ),
        (
            "validate_instrument_access/some_instrument",
            "cookie",
            200,
            False,
        ),
        (
            "validate_token",
            "password",
            200,
            False,
        ),
        (
            "validate_visit_access/some_visit",
            "cookie",
            400,
            True,
        ),
        (
            "validate_instrument_access/some_instrument",
            "password",
            400,
            True,
        ),
    ),
)
@pytest.mark.asyncio
async def test_submit_to_auth_endpoint(
    mocker: MockerFixture,
    test_params: tuple[str, str, int, bool],
):
    # Unpack test params
    url_subpath, auth_type, status_code, validation_outcome = test_params

    # Patch the auth URL to use
    auth_url = "some_url"
    mocker.patch("murfey.server.api.auth.auth_url", auth_url)

    # Patch the security config
    mock_security_config = MagicMock()
    mock_security_config.auth_url = auth_url
    mock_security_config.auth_type = auth_type
    mock_security_config.cookie_key = "_oauth2_proxy"
    mocker.patch("murfey.server.api.auth.security_config", mock_security_config)

    # Mock the request being forwarded and its headers and cookies
    mock_headers = {
        "authorization": "Bearer dummy",
        "x-auth-request-access-token": "dummy",
    }
    mock_token = "123456"
    mock_cookies = (
        {mock_security_config.cookie_key: mock_token} if auth_type == "cookie" else {}
    )

    mock_request = MagicMock()
    mock_request.headers = mock_headers

    # Mock the async response
    mock_response = MagicMock()
    mock_response.status = status_code
    mock_response.json = AsyncMock(
        return_value={
            "valid": validation_outcome,
        }
    )

    # Mock the async session and the 'get'
    mock_get = AsyncMock()
    mock_get.__aenter__.return_value = mock_response

    mock_session = MagicMock()
    mock_session.get.return_value = mock_get

    mock_session_context = AsyncMock()
    mock_session_context.__aenter__.return_value = mock_session

    mock_client_session = mocker.patch(
        "murfey.server.api.auth.aiohttp.ClientSession",
        return_value=mock_session_context,
    )

    # Run the function and check that the correct calls were made
    result = await submit_to_auth_endpoint(
        url_subpath=url_subpath,
        request=mock_request,
        token=mock_token,
    )

    # Check that aiohttp.ClientSession got called with the correct parameters
    mock_client_session.assert_called_once_with(cookies=mock_cookies)

    # Compare the headers passed to 'session.get' against what is expected
    updated_headers = copy.deepcopy(mock_headers)
    if auth_type == "password":
        updated_headers["authorization"] = f"Bearer {mock_token}"
    mock_session.get.assert_called_once_with(
        f"{mock_security_config.auth_url}/{url_subpath}",
        headers=updated_headers,
    )

    # Check that the combination of status code and JSON response are correct
    assert result == {"valid": (validation_outcome if status_code == 200 else False)}


@pytest.mark.parametrize(
    "test_params",
    (  # Exception raised? | Auth URL | Auth type | Validation outcome | User decoded | User exists
        # These cases will pass
        (False, "some_url", "cookie", True, True, True),
        (False, "", "password", True, True, True),
        # These cases will fail
        # Auth endpoint returns False
        (True, "some_url", "cookie", False, True, True),
        # Authenticating with cookie, but no auth URL
        (True, "", "cookie", True, True, True),
        # Decoding fails
        (True, "", "password", True, False, True),
        # User check fails
        (True, "", "password", True, True, False),
    ),
)
@pytest.mark.asyncio
async def test_validate_token(
    mocker: MockerFixture,
    test_params: tuple[bool, str, str, bool, bool, bool],
):
    # Unpack test params
    (
        raises_exception,
        auth_url,
        auth_type,
        validation_outcome,
        user_decoded,
        user_exists,
    ) = test_params

    # Patch the auth URL to use
    mocker.patch("murfey.server.api.auth.auth_url", auth_url)

    # Create a mock token
    mock_token = "some_token"

    # Mock the request
    mock_request = MagicMock()

    # Mock the secret key and algorithms module-level variables
    mock_secret_key = mocker.patch(
        "murfey.server.api.auth.SECRET_KEY", secrets.token_hex(32)
    )
    mock_algorithms = mocker.patch("murfey.server.api.auth.ALGORITHM", "HS256")

    # Mock the 'jwt.decode' function
    mock_decoded_data = {"user": "some_user"} if user_decoded else {}
    mock_decode = mocker.patch(
        "murfey.server.api.auth.jwt.decode", return_value=mock_decoded_data
    )

    # Mock the 'check_user' function
    mock_check_user = mocker.patch(
        "murfey.server.api.auth.check_user", return_value=user_exists
    )

    # Patch the security config
    mock_security_config = MagicMock()
    mock_security_config.auth_type = auth_type
    mocker.patch("murfey.server.api.auth.security_config", mock_security_config)

    # Patch the 'submit_to_auth_endpoint' function
    mock_submit = mocker.patch(
        "murfey.server.api.auth.submit_to_auth_endpoint", new_callable=AsyncMock
    )
    mock_submit.return_value = {"valid": validation_outcome}

    # Run the function and check that the values passed and returned are as expected
    if not raises_exception:
        result = await validate_token(
            token=mock_token,
            request=mock_request,
        )
        if auth_url:
            mock_submit.assert_called_once_with(
                "validate_token", mock_request, mock_token
            )
        if auth_type == "password":
            mock_decode.assert_called_once_with(
                mock_token, mock_secret_key, algorithms=[mock_algorithms]
            )
            mock_check_user.assert_called_once_with(mock_decoded_data["user"])
        assert result is None
    else:
        with pytest.raises(HTTPException):
            await validate_token(
                token=mock_token,
                request=mock_request,
            )


@pytest.mark.parametrize(
    "test_params",
    (  # Session ID | Visit | Expected result
        (11, "test_visit", True),
        (11, "some_visit", False),
        (12, "test_visit", False),
    ),
)
def test_validate_session_against_visit(
    mocker: MockerFixture,
    murfey_db_session: SQLModelSession,
    test_params: tuple[int, str, bool],
):
    # Unpack test params
    session_id, visit_name, expected_result = test_params

    # Add a test session to the database
    session_entry = MurfeySession(
        id=11,
        name="test_visit",
        visit="test_visit",
        started=False,
        current_gain_ref="/path/to/gain_ref",
        instrument_name="test_instrument",
        process=True,
        visit_end_time=None,
    )
    murfey_db_session.add(session_entry)
    murfey_db_session.commit()

    # Patch the Session call with the test database
    mock_session_context = MagicMock()
    mock_session_context.__enter__.return_value = murfey_db_session
    mock_session_context.__exit__.return_value = None
    mocker.patch("murfey.server.api.auth.Session", return_value=mock_session_context)

    # Run the function
    assert (
        validate_session_against_visit(
            session_id=session_id,
            visit=visit_name,
        )
        == expected_result
    )


@pytest.mark.asyncio
async def test_validate_instrument_token():
    pass


def test_get_visit_name(
    mocker: MockerFixture,
    murfey_db_session: SQLModelSession,
):
    # Add a test visit to the database
    session_entry = MurfeySession(
        id=11,
        name="test_visit",
        visit="test_visit",
        started=True,
        current_gain_ref="/path/to/gain_ref",
        instrument_name="test_instrument",
        process=True,
        visit_end_time=None,
    )
    murfey_db_session.add(session_entry)
    murfey_db_session.commit()

    # Patch the Session call with the test database
    mock_session_context = MagicMock()
    mock_session_context.__enter__.return_value = murfey_db_session
    mock_session_context.__exit__.return_value = None
    mocker.patch("murfey.server.api.auth.Session", return_value=mock_session_context)

    # Check that the built-in default visit gets returned
    assert get_visit_name(session_id=11) == "test_visit"


@pytest.mark.asyncio
async def test_validate_instrument_server_session_access():
    pass


@pytest.mark.parametrize(
    "test_params",
    (  # Raises exception | Auth URL | Validation outcome
        # These cases will pass
        (False, "some_url", True),
        (False, "", True),
        # These cases will fail
        (True, "some_url", False),
    ),
)
@pytest.mark.asyncio
async def test_validate_frontend_session_access(
    mocker: MockerFixture,
    test_params: tuple[bool, str, bool],
):
    # Unpack the test parameters
    raises_exception, auth_url, validation_outcome = test_params
    session_id = 1

    # Mock the request and token
    mock_request = MagicMock()
    mock_token = "123456"

    # Mock the results of 'get_visit_name'
    visit_name = "test_visit"
    mock_get_visit_name = mocker.patch(
        "murfey.server.api.auth.get_visit_name", return_value=visit_name
    )

    # Patch the auth URL
    mocker.patch("murfey.server.api.auth.auth_url", auth_url)

    # Patch the 'submit_to_auth_endpoint' function
    mock_submit = mocker.patch(
        "murfey.server.api.auth.submit_to_auth_endpoint", new_callable=AsyncMock
    )
    mock_submit.return_value = {"valid": validation_outcome}

    # Run the function and check that the results and passed parameters are as expected
    if not raises_exception:
        result = await validate_frontend_session_access(
            session_id=session_id,
            request=mock_request,
            token=mock_token,
        )
        mock_get_visit_name.assert_called_once_with(session_id)
        if auth_url:
            mock_submit.assert_awaited_once_with(
                f"validate_visit_access/{visit_name}",
                mock_request,
                mock_token,
            )
        else:
            mock_submit.assert_not_called()
        assert result == session_id
    else:
        with pytest.raises(HTTPException):
            await validate_frontend_session_access(
                session_id=session_id,
                request=mock_request,
                token=mock_token,
            )


@pytest.mark.parametrize(
    "test_params",
    (  # Raises exception | Auth URL | Validation outcome
        # These cases will pass
        (False, "some_url", True),
        (False, "", True),
        # These cases will fail
        (True, "some_url", False),
    ),
)
@pytest.mark.asyncio
async def test_validate_user_instrument_access(
    mocker: MockerFixture,
    test_params: tuple[bool, str, bool],
):
    # Unpack the test parameters
    raises_exception, auth_url, validation_outcome = test_params
    instrument_name = "some_instrument"

    # Mock the request and token
    mock_request = MagicMock()
    mock_token = "123456"

    # Patch the auth URL
    mocker.patch("murfey.server.api.auth.auth_url", auth_url)

    # Patch the 'submit_to_auth_endpoint' function
    mock_submit = mocker.patch(
        "murfey.server.api.auth.submit_to_auth_endpoint", new_callable=AsyncMock
    )
    mock_submit.return_value = {"valid": validation_outcome}

    # Run the function and check that the results and passed parameters are as expected
    if not raises_exception:
        result = await validate_user_instrument_access(
            instrument_name=instrument_name,
            request=mock_request,
            token=mock_token,
        )
        if auth_url:
            mock_submit.assert_awaited_once_with(
                f"validate_instrument_access/{instrument_name}",
                mock_request,
                mock_token,
            )
        else:
            mock_submit.assert_not_called()
        assert result == instrument_name
    else:
        with pytest.raises(HTTPException):
            await validate_user_instrument_access(
                instrument_name=instrument_name,
                request=mock_request,
                token=mock_token,
            )


def test_verify_password():
    pass


@pytest.mark.parametrize(
    "test_params",
    (  # User to query | Expected outcome
        ("test_user", True),
        ("some_user", False),
    ),
)
def test_validate_user(
    mocker: MockerFixture,
    murfey_db_session: SQLModelSession,
    test_params: tuple[str, bool],
):
    # Unpack test params
    user_to_query, expected_result = test_params

    # Add a user to the test database
    user_entry = MurfeyUser(
        username="test_user",
        hashed_password="asdfghjkl",
    )
    murfey_db_session.add(user_entry)
    murfey_db_session.commit()

    # Mock the 'verify_password' function
    mocker.patch("murfey.server.api.auth.verify_password", return_value=True)

    # Patch the Session call with the test database
    mock_session_context = MagicMock()
    mock_session_context.__enter__.return_value = murfey_db_session
    mock_session_context.__exit__.return_value = None
    mocker.patch("murfey.server.api.auth.Session", return_value=mock_session_context)

    # Run the function and check that the outocome is as expected
    assert validate_user(user_to_query, "dummypassword") == expected_result


def test_create_access_token():
    pass


@pytest.mark.asyncio
async def test_generate_token():
    pass


@pytest.mark.asyncio
async def test_mint_session_token():
    pass



from pathlib import Path
from unittest.mock import MagicMock

import numpy as np
import pytest
from pytest_mock import MockerFixture
from werkzeug.utils import secure_filename

from murfey.workflows.spa.atlas import atlas_jpg_from_mrc

atlas_jpg_from_mrc_test_matrix = (
    ("Atlas1.mrc",),
    ("Sample1/Atlas1.mrc",),
)


@pytest.mark.parametrize("test_params", atlas_jpg_from_mrc_test_matrix)
def test_atlas_jpg_from_mrc(
    mocker: MockerFixture, tmp_path: Path, test_params: tuple[str]
):
    # Unpack test params
    (file_name_stub,) = test_params

    # Set up mock session params
    visit_name = "test_visit"
    instrument_name = "test"
    processed_dir_name = "processed"

    # Create a 16-bit grayscale image
    shape = (64, 64)
    test_data = np.ones(shape).astype("uint16")

    # Mock out the data returned from openning the file
    mock_mrcfile = mocker.patch("murfey.workflows.spa.atlas.mrcfile")
    mock_mrc = MagicMock()
    mock_mrc.data = test_data
    mock_mrcfile.open.return_value.__enter__.return_value = mock_mrc

    # Mock the return result of 'get_machine_config()'
    mock_machine_config = MagicMock()
    mock_machine_config.processed_directory_name = processed_dir_name
    mocker.patch(
        "murfey.workflows.spa.atlas.get_machine_config",
        return_value={"test": mock_machine_config},
    )

    # Create a test file
    test_file = (
        tmp_path / instrument_name / "data" / visit_name / "atlas" / file_name_stub
    )
    test_file.parent.mkdir(parents=True, exist_ok=True)
    test_file.touch(exist_ok=True)

    # Create the expected destination directory and file
    processed_dir = (
        tmp_path / instrument_name / "data" / visit_name / processed_dir_name / "atlas"
    )
    sample_id = "Sample"
    for part in file_name_stub.split("/"):
        if part.startswith("Sample"):
            sample_id = part
            break
    processed_file_name = processed_dir / secure_filename(
        f"{sample_id}_{test_file.stem}_fullres.jpg"
    )

    # Run the function and check that the expected calls are made
    atlas_jpg_from_mrc(instrument_name, visit_name, test_file)
    assert processed_file_name.exists()



from pathlib import Path
from typing import Any
from unittest.mock import MagicMock

import ispyb.sqlalchemy as ISPyBDB
import pytest
from pytest_mock import MockerFixture
from sqlalchemy import select as sa_select
from sqlalchemy.orm.session import Session as SQLAlchemySession
from sqlmodel import select as sm_select
from sqlmodel.orm.session import Session as SQLModelSession

import murfey.util.db as MurfeyDB
from murfey.workflows.clem.register_preprocessing_results import (
    COLOR_FLAGS_MURFEY_TO_ISPYB,
    _determine_collection_mode,
    _get_color_flags,
    _register_clem_image_series,
    _register_dcg_and_atlas,
    _register_grid_square,
    _snake_to_camel_case,
    run,
)
from tests.conftest import ExampleVisit, get_or_create_db_entry

visit_name = f"{ExampleVisit.proposal_code}{ExampleVisit.proposal_number}-{ExampleVisit.visit_number}"
processed_dir_name = "processed"
grid_name = "Grid_1"


@pytest.fixture
def rsync_basepath(tmp_path: Path):
    return tmp_path / "data"


def generate_preprocessing_messages(
    rsync_basepath: Path,
    session_id: int,
    colors: list[str],
):
    # Make directory to where data for current grid is stored
    visit_dir = rsync_basepath / "2020" / visit_name
    processed_dir = visit_dir / processed_dir_name
    grid_dir = processed_dir / grid_name
    grid_dir.mkdir(parents=True, exist_ok=True)

    # Construct all the datasets to be tested
    datasets: list[tuple[Path, bool, bool, tuple[int, int], float, list[float]]] = [
        (
            grid_dir / "Overview_1" / "Image_1",
            False,
            True,
            (2400, 2400),
            1e-6,
            [0.002, 0.0044, 0.002, 0.0044],
        )
    ]
    # Add on metadata for a few grid squares
    datasets.extend(
        [
            (
                grid_dir / "TileScan_1" / f"Position_{n}",
                True,
                False,
                (2048, 2048),
                1.6e-7,
                [0.003, 0.00332768, 0.003, 0.00332768],
            )
            for n in range(5)
        ]
    )

    messages: list[dict[str, Any]] = []
    for dataset in datasets:
        # Unpack items from list of dataset parameters
        series_path = dataset[0]
        series_name = str(series_path.relative_to(processed_dir)).replace("/", "--")
        metadata = series_path / "metadata" / f"{series_path.stem}.xml"
        metadata.parent.mkdir(parents=True, exist_ok=True)
        metadata.touch(exist_ok=True)
        output_files = {color: str(series_path / f"{color}.tiff") for color in colors}
        for output_file in output_files.values():
            Path(output_file).touch(exist_ok=True)
        thumbnails = {
            color: str(series_path / ".thumbnails" / f"{color}.png") for color in colors
        }
        for v in thumbnails.values():
            if not (thumbnail := Path(v)).parent.exists():
                thumbnail.parent.mkdir(parents=True)
            thumbnail.touch(exist_ok=True)
        thumbnail_size = (512, 512)
        is_stack = dataset[1]
        is_montage = dataset[2]
        shape = dataset[3]
        pixel_size = dataset[4]
        extent = dataset[5]

        message = {
            "session_id": session_id,
            "result": {
                "series_name": series_name,
                "number_of_members": 3,
                "is_stack": is_stack,
                "is_montage": is_montage,
                "output_files": output_files,
                "thumbnails": thumbnails,
                "thumbnail_size": thumbnail_size,
                "metadata": str(metadata),
                "parent_lif": None,
                "parent_tiffs": {},
                "pixels_x": shape[0],
                "pixels_y": shape[1],
                "units": "m",
                "pixel_size": pixel_size,
                "resolution": 1 / pixel_size,
                "extent": extent,
            },
        }
        messages.append(message)
    return messages


@pytest.mark.parametrize(
    "test_params",
    (
        (
            ["gray"],
            {
                "has_grey": True,
            },
        ),
        (
            ["gray", "red"],
            {
                "has_grey": True,
                "has_red": True,
            },
        ),
        (
            ["red", "green", "blue"],
            {
                "has_red": True,
                "has_green": True,
                "has_blue": True,
            },
        ),
        (
            ["cyan", "magenta", "yellow"],
            {
                "has_cyan": True,
                "has_magenta": True,
                "has_yellow": True,
            },
        ),
    ),
)
def test_get_color_flags(test_params: tuple[list[str], dict[str, bool]]):
    colors, positive_flags = test_params
    expected_result = dict.fromkeys(
        (
            "has_grey",
            "has_red",
            "has_green",
            "has_blue",
            "has_cyan",
            "has_magenta",
            "has_yellow",
        ),
        False,
    )
    for flag, value in positive_flags.items():
        expected_result[flag] = value
    assert _get_color_flags(colors) == expected_result


def test_register_clem_image_series():
    _register_clem_image_series


@pytest.mark.parametrize(
    "test_params",
    (
        (["gray"], "Bright Field"),
        (["gray", "blue"], "Bright Field and Fluorescent"),
        (["red", "green", "blue"], "Fluorescent"),
    ),
)
def test_determine_collection_mode(test_params: tuple[list[str], str]):
    colors, expected_result = test_params
    assert _determine_collection_mode(colors) == expected_result


@pytest.mark.parametrize(
    "test_params",
    (
        ("has_grey", "hasGrey"),
        ("has_red", "hasRed"),
        ("has_green", "hasGreen"),
        ("has_blue", "hasBlue"),
        ("has_cyan", "hasCyan"),
        ("has_magenta", "hasMagenta"),
        ("has_yellow", "hasYellow"),
    ),
)
def test_snake_to_camel_case(
    test_params: tuple[str, str],
):
    string, expected_result = test_params
    assert _snake_to_camel_case(string) == expected_result


def test_register_dcg_and_atlas():
    _register_dcg_and_atlas


def test_register_grid_square():
    _register_grid_square


@pytest.mark.parametrize(
    "test_params",
    (  # Colors
        (["gray"],),
        (["gray", "green"],),
        (["red", "green", "blue"],),
        (["cyan", "magenta", "blue"],),
    ),
)
def test_run(
    mocker: MockerFixture,
    rsync_basepath: Path,
    test_params: tuple[list[str]],
):
    # Unpack test params
    (colors,) = test_params

    # Mock the MurfeyDB connection
    mock_murfey_session_entry = MagicMock()
    mock_murfey_session_entry.instrument_name = ExampleVisit.instrument_name
    mock_murfey_session_entry.visit = visit_name
    mock_murfey_db = MagicMock()
    mock_murfey_db.exec().return_value.one.return_value = mock_murfey_session_entry

    # Mock the registration helper functions
    mock_register_clem_series = mocker.patch(
        "murfey.workflows.clem.register_preprocessing_results._register_clem_image_series"
    )
    mock_register_dcg_and_atlas = mocker.patch(
        "murfey.workflows.clem.register_preprocessing_results._register_dcg_and_atlas"
    )
    mock_register_grid_square = mocker.patch(
        "murfey.workflows.clem.register_preprocessing_results._register_grid_square"
    )

    # Mock the align and merge workflow call
    mock_align_and_merge_call = mocker.patch(
        "murfey.workflows.clem.register_preprocessing_results.submit_cluster_request"
    )

    preprocessing_messages = generate_preprocessing_messages(
        rsync_basepath=rsync_basepath,
        session_id=ExampleVisit.murfey_session_id,
        colors=colors,
    )
    for message in preprocessing_messages:
        result = run(
            message=message,
            murfey_db=mock_murfey_db,
        )
        assert result == {"success": True}
    assert mock_register_clem_series.call_count == len(preprocessing_messages)
    assert mock_register_dcg_and_atlas.call_count == len(preprocessing_messages)
    assert mock_register_grid_square.call_count == len(preprocessing_messages)
    if ("gray" not in colors) or ("gray" in colors and len(colors) == 1):
        assert mock_align_and_merge_call.call_count == len(preprocessing_messages)
    else:
        assert mock_align_and_merge_call.call_count == len(preprocessing_messages) * 3


@pytest.mark.parametrize(
    "test_params",
    (
        # Reverse list order? | Colors
        (False, ["gray"]),
        (True, ["gray"]),
        (False, ["red", "green", "blue"]),
        (True, ["cyan", "magenta", "yellow"]),
        (False, ["gray", "red", "green", "blue"]),
        (True, ["gray", "cyan", "magenta", "yellow"]),
    ),
)
def test_run_with_db(
    mocker: MockerFixture,
    rsync_basepath: Path,
    mock_ispyb_credentials,
    murfey_db_session: SQLModelSession,
    ispyb_db_session: SQLAlchemySession,
    test_params: tuple[bool, list[str]],
):
    # Unpack test params
    (shuffle_message, colors) = test_params

    # Create a session to insert for this test
    murfey_session: MurfeyDB.Session = get_or_create_db_entry(
        murfey_db_session,
        MurfeyDB.Session,
        lookup_kwargs={
            "id": ExampleVisit.murfey_session_id + 1,
            "name": visit_name,
            "visit": visit_name,
            "instrument_name": ExampleVisit.instrument_name,
        },
    )

    # Mock the ISPyB connection where the TransportManager class is located
    mock_security_config = MagicMock()
    mock_security_config.ispyb_credentials = mock_ispyb_credentials
    mocker.patch(
        "murfey.server.ispyb.get_security_config",
        return_value=mock_security_config,
    )
    mocker.patch(
        "murfey.server.ispyb.ISPyBSession",
        return_value=ispyb_db_session,
    )

    # Mock the ISPYB connection when registering data collection group
    mocker.patch(
        "murfey.workflows.register_data_collection_group.ISPyBSession",
        return_value=ispyb_db_session,
    )

    # Mock out the machine config used in the helper sanitisation function
    mock_get_machine_config = mocker.patch("murfey.workflows.clem.get_machine_config")
    mock_machine_config = MagicMock()
    mock_machine_config.rsync_basepath = rsync_basepath
    mock_get_machine_config.return_value = {
        ExampleVisit.instrument_name: mock_machine_config,
    }

    # Mock the align and merge workflow call
    mock_align_and_merge_call = mocker.patch(
        "murfey.workflows.clem.register_preprocessing_results.submit_cluster_request"
    )

    # Patch the TransportManager object in the workflows called
    from murfey.server.ispyb import TransportManager

    mocker.patch(
        "murfey.workflows.clem.register_preprocessing_results._transport_object",
        new=TransportManager("PikaTransport"),
    )
    mocker.patch(
        "murfey.workflows.register_data_collection_group._transport_object",
        new=TransportManager("PikaTransport"),
    )
    mocker.patch(
        "murfey.workflows.register_atlas_update._transport_object",
        new=TransportManager("PikaTransport"),
    )

    # Run the function
    preprocessing_messages = generate_preprocessing_messages(
        rsync_basepath=rsync_basepath,
        session_id=murfey_session.id,
        colors=colors,
    )
    if shuffle_message:
        preprocessing_messages.reverse()
    for message in preprocessing_messages:
        result = run(
            message=message,
            murfey_db=murfey_db_session,
        )
        assert result == {"success": True}

    # Each message should call the align-and-merge workflow thrice
    # if gray and colour channels are both present
    if ("gray" not in colors) or ("gray" in colors and len(colors) == 1):
        assert mock_align_and_merge_call.call_count == len(preprocessing_messages)
    else:
        assert mock_align_and_merge_call.call_count == len(preprocessing_messages) * 3

    # Both databases should have entries for data collection group, and grid squares
    # ISPyB database should additionally have an atlas entry
    murfey_dcg_search = murfey_db_session.exec(
        sm_select(MurfeyDB.DataCollectionGroup).where(
            MurfeyDB.DataCollectionGroup.session_id == murfey_session.id
        )
    ).all()
    assert len(murfey_dcg_search) == 1
    murfey_gs_search = murfey_db_session.exec(
        sm_select(MurfeyDB.GridSquare).where(
            MurfeyDB.GridSquare.session_id == murfey_session.id
        )
    ).all()
    assert len(murfey_gs_search) == len(preprocessing_messages) - 1

    murfey_dcg = murfey_dcg_search[0]
    ispyb_dcg_search = (
        ispyb_db_session.execute(
            sa_select(ISPyBDB.DataCollectionGroup).where(
                ISPyBDB.DataCollectionGroup.dataCollectionGroupId == murfey_dcg.id
            )
        )
        .scalars()
        .all()
    )
    assert len(ispyb_dcg_search) == 1

    ispyb_dcg = ispyb_dcg_search[0]
    ispyb_atlas_search = (
        ispyb_db_session.execute(
            sa_select(ISPyBDB.Atlas).where(
                ISPyBDB.Atlas.dataCollectionGroupId == ispyb_dcg.dataCollectionGroupId
            )
        )
        .scalars()
        .all()
    )
    assert len(ispyb_atlas_search) == 1

    # Determine the color flags and collection mode
    color_flags = {
        COLOR_FLAGS_MURFEY_TO_ISPYB[flag]: int(value)
        for flag, value in _get_color_flags(colors).items()
    }
    collection_mode = _determine_collection_mode(colors)

    ispyb_atlas = ispyb_atlas_search[0]
    # Check that the Atlas color flags and collection mode are set correctly
    for flag, value in color_flags.items():
        assert getattr(ispyb_atlas, flag) == value
    assert ispyb_atlas.mode == collection_mode

    ispyb_gs_search = (
        ispyb_db_session.execute(
            sa_select(ISPyBDB.GridSquare).where(
                ISPyBDB.GridSquare.atlasId == ispyb_atlas.atlasId
            )
        )
        .scalars()
        .all()
    )
    assert len(ispyb_gs_search) == len(preprocessing_messages) - 1
    for gs in ispyb_gs_search:
        # Check that the Atlas color flags and collection mode are set correctly
        for flag, value in color_flags.items():
            assert getattr(gs, flag) == value
        assert gs.mode == collection_mode

    murfey_db_session.close()
    ispyb_db_session.close()
